import numpy as np
from celldetective.io import auto_load_number_of_frames, load_frames
from celldetective.filters import *
from celldetective.segmentation import filter_image, threshold_image
from celldetective.measure import contour_of_instance_segmentation
from celldetective.utils import _get_img_num_per_channel, estimate_unreliable_edge
from tifffile import imread
import matplotlib.pyplot as plt 
from pathlib import Path
from natsort import natsorted
from glob import glob
import os

from PyQt5.QtWidgets import QWidget, QHBoxLayout, QPushButton, QLabel, QComboBox, QLineEdit, QListWidget
from PyQt5.QtCore import Qt, QSize
from celldetective.gui.gui_utils import FigureCanvas, center_window, QuickSliderLayout
from celldetective.gui import Styles
from superqt import QLabeledDoubleSlider, QLabeledSlider, QLabeledDoubleRangeSlider
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
from matplotlib_scalebar.scalebar import ScaleBar
import gc
from celldetective.utils import mask_edges

class StackVisualizer(QWidget, Styles):

	"""
	A widget for visualizing image stacks with interactive sliders and channel selection.

	Parameters:
	- stack (numpy.ndarray or None): The stack of images.
	- stack_path (str or None): The path to the stack of images if provided as a file.
	- frame_slider (bool): Enable frame navigation slider.
	- contrast_slider (bool): Enable contrast adjustment slider.
	- channel_cb (bool): Enable channel selection dropdown.
	- channel_names (list or None): Names of the channels if `channel_cb` is True.
	- n_channels (int): Number of channels.
	- target_channel (int): Index of the target channel.
	- window_title (str): Title of the window.
	- PxToUm (float or None): Pixel to micrometer conversion factor.
	- background_color (str): Background color of the widget.
	- imshow_kwargs (dict): Additional keyword arguments for imshow function.

	Methods:
	- show(): Display the widget.
	- load_stack(): Load the stack of images.
	- locate_image_virtual(): Locate the stack of images if provided as a file.
	- generate_figure_canvas(): Generate the figure canvas for displaying images.
	- generate_channel_cb(): Generate the channel dropdown if enabled.
	- generate_contrast_slider(): Generate the contrast slider if enabled.
	- generate_frame_slider(): Generate the frame slider if enabled.
	- set_target_channel(value): Set the target channel.
	- change_contrast(value): Change contrast based on slider value.
	- set_channel_index(value): Set the channel index based on dropdown value.
	- change_frame(value): Change the displayed frame based on slider value.
	- closeEvent(event): Event handler for closing the widget.

	Notes:
	- This class provides a convenient interface for visualizing image stacks with frame navigation,
	  contrast adjustment, and channel selection functionalities.
	"""
	
	def __init__(self, stack=None, stack_path=None, frame_slider=True, contrast_slider=True, channel_cb=False, channel_names=None, n_channels=1, target_channel=0, window_title='View', PxToUm=None, background_color='transparent',imshow_kwargs={}):
		super().__init__()

		#self.setWindowTitle(window_title)
		self.window_title = window_title

		self.stack = stack
		self.stack_path = stack_path
		self.create_frame_slider = frame_slider
		self.background_color = background_color
		self.create_contrast_slider = contrast_slider
		self.create_channel_cb = channel_cb
		self.n_channels = n_channels
		self.channel_names = channel_names
		self.target_channel = target_channel
		self.imshow_kwargs = imshow_kwargs
		self.PxToUm = PxToUm
		self.init_contrast = False

		self.load_stack() # need to get stack, frame etc
		self.generate_figure_canvas()
		if self.create_channel_cb:
			self.generate_channel_cb()
		if self.create_contrast_slider:
			self.generate_contrast_slider()
		if self.create_frame_slider:
			self.generate_frame_slider()

		self.canvas.layout.setContentsMargins(15,15,15,30)
		self.setAttribute(Qt.WA_DeleteOnClose)
		center_window(self)

	def show(self):
		# Display the widget
		self.canvas.show()

	def load_stack(self):
		# Load the stack of images
		if self.stack is not None:

			if isinstance(self.stack, list):
				self.stack = np.array(self.stack)

			if self.stack.ndim==3:
				print('No channel axis found...')
				self.stack = self.stack[:,:,:,np.newaxis]
				self.target_channel = 0
			
			self.mode = 'direct'
			self.stack_length = len(self.stack)
			self.mid_time = self.stack_length // 2
			self.init_frame = self.stack[self.mid_time,:,:,self.target_channel]
			self.last_frame = self.stack[-1,:,:,self.target_channel]
		else:
			self.mode = 'virtual'
			assert isinstance(self.stack_path, str)
			assert self.stack_path.endswith('.tif')
			self.locate_image_virtual()

	def locate_image_virtual(self):
		# Locate the stack of images if provided as a file
		self.stack_length = auto_load_number_of_frames(self.stack_path)
		if self.stack_length is None:
			stack = imread(self.stack_path)
			self.stack_length = len(stack)
			del stack
			gc.collect()

		self.mid_time = self.stack_length // 2
		self.img_num_per_channel = _get_img_num_per_channel(np.arange(self.n_channels), self.stack_length, self.n_channels)

		self.init_frame = load_frames(self.img_num_per_channel[self.target_channel, self.mid_time], 
									  self.stack_path,
									  normalize_input=False).astype(float)[:,:,0]
		self.last_frame = load_frames(self.img_num_per_channel[self.target_channel, self.stack_length-1], 
									  self.stack_path,
									  normalize_input=False).astype(float)[:,:,0]

	def generate_figure_canvas(self):
		# Generate the figure canvas for displaying images

		self.fig, self.ax = plt.subplots(figsize=(5,5),tight_layout=True) #figsize=(5, 5)
		self.canvas = FigureCanvas(self.fig, title=self.window_title, interactive=True)
		self.ax.clear()
		self.im = self.ax.imshow(self.init_frame, cmap='gray', interpolation='none', **self.imshow_kwargs)
		if self.PxToUm is not None:
			scalebar = ScaleBar(self.PxToUm,
								"um",
								length_fraction=0.25,
								location='upper right',
								border_pad=0.4,
								box_alpha=0.95,
								color='white',
								box_color='black',
								)
			if self.PxToUm==1:
				scalebar = ScaleBar(1,
								"px",
								dimension="pixel-length",
								length_fraction=0.25,
								location='upper right',
								border_pad=0.4,
								box_alpha=0.95,
								color='white',
								box_color='black',
								)
			self.ax.add_artist(scalebar)
		self.ax.set_xticks([])
		self.ax.set_yticks([])
		self.fig.set_facecolor('none')  # or 'None'
		self.fig.canvas.setStyleSheet(f"background-color: {self.background_color};")
		self.canvas.canvas.draw()

	def generate_channel_cb(self):
		# Generate the channel dropdown if enabled

		assert self.channel_names is not None
		assert len(self.channel_names)==self.n_channels

		channel_layout = QHBoxLayout()
		channel_layout.setContentsMargins(15,0,15,0)
		channel_layout.addWidget(QLabel('Channel: '), 25)

		self.channels_cb = QComboBox()
		self.channels_cb.addItems(self.channel_names)
		self.channels_cb.currentIndexChanged.connect(self.set_channel_index)
		channel_layout.addWidget(self.channels_cb, 75)
		self.canvas.layout.addLayout(channel_layout)

	def generate_contrast_slider(self):
		# Generate the contrast slider if enabled
		
		self.contrast_slider = QLabeledDoubleRangeSlider()
		contrast_layout = QuickSliderLayout(
											label='Contrast: ',
											slider=self.contrast_slider,
											slider_initial_value=[np.nanpercentile(self.init_frame, 0.1),np.nanpercentile(self.init_frame, 99.99)],
											slider_range=(np.nanmin(self.init_frame),np.nanmax(self.init_frame)),
											decimal_option=True,
											precision=1.0E-05,
											)
		contrast_layout.setContentsMargins(15,0,15,0)
		self.im.set_clim(vmin=np.nanpercentile(self.init_frame, 0.1),vmax=np.nanpercentile(self.init_frame, 99.99))
		self.contrast_slider.valueChanged.connect(self.change_contrast)
		self.canvas.layout.addLayout(contrast_layout)



	def generate_frame_slider(self):
		# Generate the frame slider if enabled
	
		self.frame_slider = QLabeledSlider()
		frame_layout = QuickSliderLayout(
										label='Frame: ',
										slider=self.frame_slider,
										slider_initial_value=int(self.mid_time),
										slider_range=(0,self.stack_length-1),
										decimal_option=False,
										)
		frame_layout.setContentsMargins(15,0,15,0)
		self.frame_slider.valueChanged.connect(self.change_frame)
		self.canvas.layout.addLayout(frame_layout)

	def set_target_channel(self, value):
		# Set the target channel
		
		self.target_channel = value
		self.change_frame(self.frame_slider.value())

	def change_contrast(self, value):
		# Change contrast based on slider value

		vmin = value[0]
		vmax = value[1]
		self.im.set_clim(vmin=vmin, vmax=vmax)
		self.fig.canvas.draw_idle()

	def set_channel_index(self, value):
		# Set the channel index based on dropdown value

		self.target_channel = value
		self.init_contrast = True
		if self.mode == 'direct':
			self.last_frame = self.stack[-1,:,:,self.target_channel]
		elif self.mode == 'virtual':
			self.last_frame = load_frames(self.img_num_per_channel[self.target_channel, self.stack_length-1], 
										  self.stack_path,
										  normalize_input=False).astype(float)[:,:,0]
		self.change_frame(self.frame_slider.value())
		self.init_contrast = False

	def change_frame(self, value):
		# Change the displayed frame based on slider value
		
		if self.mode=='virtual':

			self.init_frame = load_frames(self.img_num_per_channel[self.target_channel, value], 
								self.stack_path,
								normalize_input=False
								).astype(float)[:,:,0]
		elif self.mode=='direct':
			self.init_frame = self.stack[value,:,:,self.target_channel].copy()
		
		self.im.set_data(self.init_frame)
		
		if self.init_contrast:
			self.im.autoscale()
			I_min, I_max = self.im.get_clim()
			self.contrast_slider.setRange(np.nanmin([self.init_frame,self.last_frame]),np.nanmax([self.init_frame,self.last_frame]))
			self.contrast_slider.setValue((I_min,I_max))

		if self.create_contrast_slider:
			self.change_contrast(self.contrast_slider.value())

	
	def closeEvent(self, event):
		# Event handler for closing the widget
		self.canvas.close()


class ThresholdedStackVisualizer(StackVisualizer):
	
	"""
	A widget for visualizing thresholded image stacks with interactive sliders and channel selection.

	Parameters:
	- preprocessing (list or None): A list of preprocessing filters to apply to the image before thresholding.
	- parent_le: The parent QLineEdit instance to set the threshold value.
	- initial_threshold (float): Initial threshold value.
	- initial_mask_alpha (float): Initial mask opacity value.
	- args, kwargs: Additional arguments to pass to the parent class constructor.

	Methods:
	- generate_apply_btn(): Generate the apply button to set the threshold in the parent QLineEdit.
	- set_threshold_in_parent_le(): Set the threshold value in the parent QLineEdit.
	- generate_mask_imshow(): Generate the mask imshow.
	- generate_threshold_slider(): Generate the threshold slider.
	- generate_opacity_slider(): Generate the opacity slider for the mask.
	- change_mask_opacity(value): Change the opacity of the mask.
	- change_threshold(value): Change the threshold value.
	- change_frame(value): Change the displayed frame and update the threshold.
	- compute_mask(threshold_value): Compute the mask based on the threshold value.
	- preprocess_image(): Preprocess the image before thresholding.

	Notes:
	- This class extends the functionality of StackVisualizer to visualize thresholded image stacks
	  with interactive sliders for threshold and mask opacity adjustment.
	"""

	def __init__(self, preprocessing=None, parent_le=None, initial_threshold=5, initial_mask_alpha=0.5, *args, **kwargs):
		# Initialize the widget and its attributes		
		super().__init__(*args, **kwargs)
		self.preprocessing = preprocessing
		self.thresh = initial_threshold
		self.mask_alpha = initial_mask_alpha
		self.parent_le = parent_le
		self.compute_mask(self.thresh)
		self.generate_mask_imshow()
		self.generate_threshold_slider()
		self.generate_opacity_slider()
		if isinstance(self.parent_le, QLineEdit):
			self.generate_apply_btn()

	def generate_apply_btn(self):
		# Generate the apply button to set the threshold in the parent QLineEdit		
		apply_hbox = QHBoxLayout()
		self.apply_threshold_btn = QPushButton('Apply')
		self.apply_threshold_btn.clicked.connect(self.set_threshold_in_parent_le)
		self.apply_threshold_btn.setStyleSheet(self.button_style_sheet)
		apply_hbox.addWidget(QLabel(''),33)
		apply_hbox.addWidget(self.apply_threshold_btn, 33)
		apply_hbox.addWidget(QLabel(''),33)		
		self.canvas.layout.addLayout(apply_hbox)

	def set_threshold_in_parent_le(self):
		# Set the threshold value in the parent QLineEdit
		self.parent_le.set_threshold(self.threshold_slider.value())
		self.close()

	def generate_mask_imshow(self):
		# Generate the mask imshow		
		self.im_mask = self.ax.imshow(np.ma.masked_where(self.mask==0, self.mask), alpha=self.mask_alpha, interpolation='none')
		self.canvas.canvas.draw()

	def generate_threshold_slider(self):
		# Generate the threshold slider
		self.threshold_slider = QLabeledDoubleSlider()
		thresh_layout = QuickSliderLayout(label='Threshold: ',
										slider=self.threshold_slider,
										slider_initial_value=self.thresh,
										slider_range=(0,30),
										decimal_option=True,
										precision=1.0E-05,
										)
		thresh_layout.setContentsMargins(15,0,15,0)
		self.threshold_slider.valueChanged.connect(self.change_threshold)
		self.canvas.layout.addLayout(thresh_layout)

	def generate_opacity_slider(self):
		# Generate the opacity slider for the mask
		self.opacity_slider = QLabeledDoubleSlider()
		opacity_layout = QuickSliderLayout(label='Opacity: ',
										slider=self.opacity_slider,
										slider_initial_value=0.5,
										slider_range=(0,1),
										decimal_option=True,
										precision=1.0E-03
										)
		opacity_layout.setContentsMargins(15,0,15,0)
		self.opacity_slider.valueChanged.connect(self.change_mask_opacity)
		self.canvas.layout.addLayout(opacity_layout)

	def change_mask_opacity(self, value):
		# Change the opacity of the mask
		self.mask_alpha = value
		self.im_mask.set_alpha(self.mask_alpha)
		self.canvas.canvas.draw_idle()

	def change_threshold(self, value):
		# Change the threshold value		
		self.thresh = value
		self.compute_mask(self.thresh)
		mask = np.ma.masked_where(self.mask == 0, self.mask)
		self.im_mask.set_data(mask)
		self.canvas.canvas.draw_idle()

	def change_frame(self, value):
		# Change the displayed frame and update the threshold		
		super().change_frame(value)
		self.change_threshold(self.threshold_slider.value())

	def compute_mask(self, threshold_value):
		# Compute the mask based on the threshold value
		self.preprocess_image()
		edge = estimate_unreliable_edge(self.preprocessing)
		self.mask = threshold_image(self.processed_image, threshold_value, np.inf, foreground_value=1, edge_exclusion=edge).astype(int)

	def preprocess_image(self):
		# Preprocess the image before thresholding		
		if self.preprocessing is not None:

			assert isinstance(self.preprocessing, list)
			self.processed_image = filter_image(self.init_frame.copy(),filters=self.preprocessing)


class CellEdgeVisualizer(StackVisualizer):

	"""
	A widget for visualizing cell edges with interactive sliders and channel selection.

	Parameters:
	- cell_type (str): Type of cells ('effectors' by default).
	- edge_range (tuple): Range of edge sizes (-30, 30) by default.
	- invert (bool): Flag to invert the edge size (False by default).
	- parent_list_widget: The parent QListWidget instance to add edge measurements.
	- parent_le: The parent QLineEdit instance to set the edge size.
	- labels (array or None): Array of labels for cell segmentation.
	- initial_edge (int): Initial edge size (5 by default).
	- initial_mask_alpha (float): Initial mask opacity value (0.5 by default).
	- args, kwargs: Additional arguments to pass to the parent class constructor.

	Methods:
	- load_labels(): Load the cell labels.
	- locate_labels_virtual(): Locate virtual labels.
	- generate_add_to_list_btn(): Generate the add to list button.
	- generate_add_to_le_btn(): Generate the set measurement button for QLineEdit.
	- set_measurement_in_parent_le(): Set the edge size in the parent QLineEdit.
	- set_measurement_in_parent_list(): Add the edge size to the parent QListWidget.
	- generate_label_imshow(): Generate the label imshow.
	- generate_edge_slider(): Generate the edge size slider.
	- generate_opacity_slider(): Generate the opacity slider for the mask.
	- change_mask_opacity(value): Change the opacity of the mask.
	- change_edge_size(value): Change the edge size.
	- change_frame(value): Change the displayed frame and update the edge labels.
	- compute_edge_labels(): Compute the edge labels.

	Notes:
	- This class extends the functionality of StackVisualizer to visualize cell edges
	  with interactive sliders for edge size adjustment and mask opacity control.
	"""

	def __init__(self, cell_type="effectors", edge_range=(-30,30), invert=False, parent_list_widget=None, parent_le=None, labels=None, initial_edge=5, initial_mask_alpha=0.5, *args, **kwargs):
		
		# Initialize the widget and its attributes
		super().__init__(*args, **kwargs)
		self.edge_size = initial_edge
		self.mask_alpha = initial_mask_alpha
		self.cell_type = cell_type
		self.labels = labels
		self.edge_range = edge_range
		self.invert = invert
		self.parent_list_widget = parent_list_widget
		self.parent_le = parent_le

		self.load_labels()
		self.generate_label_imshow()
		self.generate_edge_slider()
		self.generate_opacity_slider()
		if isinstance(self.parent_list_widget, QListWidget):
			self.generate_add_to_list_btn()
		if isinstance(self.parent_le, QLineEdit):
			self.generate_add_to_le_btn()

	def load_labels(self):
		# Load the cell labels

		if self.labels is not None:

			if isinstance(self.labels, list):
				self.labels = np.array(self.labels)

			assert self.labels.ndim==3,'Wrong dimensions for the provided labels, expect TXY'
			assert len(self.labels)==self.stack_length

			self.mode = 'direct'
			self.init_label = self.labels[self.mid_time,:,:]
		else:
			self.mode = 'virtual'
			assert isinstance(self.stack_path, str)
			assert self.stack_path.endswith('.tif')
			self.locate_labels_virtual()
		
		self.compute_edge_labels()

	def locate_labels_virtual(self):
		# Locate virtual labels

		labels_path = str(Path(self.stack_path).parent.parent) + os.sep + f'labels_{self.cell_type}' + os.sep
		self.mask_paths = natsorted(glob(labels_path + '*.tif'))
		
		if len(self.mask_paths) == 0:

			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Critical)
			msgBox.setText("No labels were found for the selected cells. Abort.")
			msgBox.setWindowTitle("Critical")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			self.close()

		self.init_label = imread(self.mask_paths[self.frame_slider.value()])

	def generate_add_to_list_btn(self):
		# Generate the add to list button
		
		add_hbox = QHBoxLayout()
		self.add_measurement_btn = QPushButton('Add measurement')
		self.add_measurement_btn.clicked.connect(self.set_measurement_in_parent_list)
		self.add_measurement_btn.setIcon(icon(MDI6.plus,color="white"))
		self.add_measurement_btn.setIconSize(QSize(20, 20))
		self.add_measurement_btn.setStyleSheet(self.button_style_sheet)
		add_hbox.addWidget(QLabel(''),33)
		add_hbox.addWidget(self.add_measurement_btn, 33)
		add_hbox.addWidget(QLabel(''),33)		
		self.canvas.layout.addLayout(add_hbox)

	def generate_add_to_le_btn(self):
		# Generate the set measurement button for QLineEdit
		
		add_hbox = QHBoxLayout()
		self.set_measurement_btn = QPushButton('Set')
		self.set_measurement_btn.clicked.connect(self.set_measurement_in_parent_le)
		self.set_measurement_btn.setStyleSheet(self.button_style_sheet)
		add_hbox.addWidget(QLabel(''),33)
		add_hbox.addWidget(self.set_measurement_btn, 33)
		add_hbox.addWidget(QLabel(''),33)		
		self.canvas.layout.addLayout(add_hbox)

	def set_measurement_in_parent_le(self):
		# Set the edge size in the parent QLineEdit
		
		self.parent_le.setText(str(int(self.edge_slider.value())))
		self.close()

	def set_measurement_in_parent_list(self):
		# Add the edge size to the parent QListWidget
		
		self.parent_list_widget.addItems([str(self.edge_slider.value())])
		self.close()

	def generate_label_imshow(self):
		# Generate the label imshow
		
		self.im_mask = self.ax.imshow(np.ma.masked_where(self.edge_labels==0, self.edge_labels), alpha=self.mask_alpha, interpolation='none', cmap="viridis")
		self.canvas.canvas.draw()

	def generate_edge_slider(self):
		# Generate the edge size slider

		self.edge_slider = QLabeledSlider()
		edge_layout = QuickSliderLayout(label='Edge: ',
										slider=self.edge_slider,
										slider_initial_value=self.edge_size,
										slider_range=self.edge_range,
										decimal_option=False,
										)
		edge_layout.setContentsMargins(15,0,15,0)
		self.edge_slider.valueChanged.connect(self.change_edge_size)
		self.canvas.layout.addLayout(edge_layout)

	def generate_opacity_slider(self):
		# Generate the opacity slider for the mask

		self.opacity_slider = QLabeledDoubleSlider()
		opacity_layout = QuickSliderLayout(label='Opacity: ',
										slider=self.opacity_slider,
										slider_initial_value=0.5,
										slider_range=(0,1),
										decimal_option=True,
										precision=1.0E-03
										)
		opacity_layout.setContentsMargins(15,0,15,0)
		self.opacity_slider.valueChanged.connect(self.change_mask_opacity)
		self.canvas.layout.addLayout(opacity_layout)

	def change_mask_opacity(self, value):
		# Change the opacity of the mask

		self.mask_alpha = value
		self.im_mask.set_alpha(self.mask_alpha)
		self.canvas.canvas.draw_idle()

	def change_edge_size(self, value):
		# Change the edge size
		
		self.edge_size = value
		self.compute_edge_labels()
		mask = np.ma.masked_where(self.edge_labels == 0, self.edge_labels)
		self.im_mask.set_data(mask)
		self.canvas.canvas.draw_idle()

	def change_frame(self, value):
		# Change the displayed frame and update the edge labels
		
		super().change_frame(value)

		if self.mode=='virtual':
			self.init_label = imread(self.mask_paths[value])
		elif self.mode=='direct':
			self.init_label = self.labels[value,:,:]
		
		self.compute_edge_labels()
		mask = np.ma.masked_where(self.edge_labels == 0, self.edge_labels)
		self.im_mask.set_data(mask)

	def compute_edge_labels(self):
		# Compute the edge labels
		
		if self.invert:
			edge_size = - self.edge_size
		else:
			edge_size = self.edge_size

		self.edge_labels = contour_of_instance_segmentation(self.init_label, edge_size)

class CellSizeViewer(StackVisualizer):

	"""
	A widget for visualizing cell size with interactive sliders and circle display.

	Parameters:
	- initial_diameter (int): Initial diameter of the circle (40 by default).
	- set_radius_in_list (bool): Flag to set radius instead of diameter in the list (False by default).
	- diameter_slider_range (tuple): Range of the diameter slider (0, 200) by default.
	- parent_le: The parent QLineEdit instance to set the diameter.
	- parent_list_widget: The parent QListWidget instance to add diameter measurements.
	- args, kwargs: Additional arguments to pass to the parent class constructor.

	Methods:
	- generate_circle(): Generate the circle for visualization.
	- generate_add_to_list_btn(): Generate the add to list button.
	- set_measurement_in_parent_list(): Add the diameter to the parent QListWidget.
	- on_xlims_or_ylims_change(event_ax): Update the circle position on axis limits change.
	- generate_set_btn(): Generate the set button for QLineEdit.
	- set_threshold_in_parent_le(): Set the diameter in the parent QLineEdit.
	- generate_diameter_slider(): Generate the diameter slider.
	- change_diameter(value): Change the diameter of the circle.

	Notes:
	- This class extends the functionality of StackVisualizer to visualize cell size
	  with interactive sliders for diameter adjustment and circle display.
	"""

	def __init__(self, initial_diameter=40, set_radius_in_list=False, diameter_slider_range=(0,200), parent_le=None, parent_list_widget=None, *args, **kwargs):
		# Initialize the widget and its attributes
		
		super().__init__(*args, **kwargs)
		self.diameter = initial_diameter
		self.parent_le = parent_le
		self.diameter_slider_range = diameter_slider_range
		self.parent_list_widget = parent_list_widget
		self.set_radius_in_list = set_radius_in_list
		self.generate_circle()
		self.generate_diameter_slider()

		if isinstance(self.parent_le, QLineEdit):
			self.generate_set_btn()
		if isinstance(self.parent_list_widget, QListWidget):
			self.generate_add_to_list_btn()

	def generate_circle(self):
		# Generate the circle for visualization

		self.circ = plt.Circle((self.init_frame.shape[0]//2,self.init_frame.shape[1]//2), self.diameter//2, ec="tab:red",fill=False)
		self.ax.add_patch(self.circ)

		self.ax.callbacks.connect('xlim_changed',self.on_xlims_or_ylims_change)
		self.ax.callbacks.connect('ylim_changed', self.on_xlims_or_ylims_change)

	def generate_add_to_list_btn(self):
		# Generate the add to list button
		
		add_hbox = QHBoxLayout()
		self.add_measurement_btn = QPushButton('Add measurement')
		self.add_measurement_btn.clicked.connect(self.set_measurement_in_parent_list)
		self.add_measurement_btn.setIcon(icon(MDI6.plus,color="white"))
		self.add_measurement_btn.setIconSize(QSize(20, 20))
		self.add_measurement_btn.setStyleSheet(self.button_style_sheet)
		add_hbox.addWidget(QLabel(''),33)
		add_hbox.addWidget(self.add_measurement_btn, 33)
		add_hbox.addWidget(QLabel(''),33)		
		self.canvas.layout.addLayout(add_hbox)

	def set_measurement_in_parent_list(self):
		# Add the diameter to the parent QListWidget

		if self.set_radius_in_list:
			val = int(self.diameter_slider.value()//2)
		else:
			val = int(self.diameter_slider.value())
		
		self.parent_list_widget.addItems([str(val)])
		self.close()

	def on_xlims_or_ylims_change(self, event_ax):
		# Update the circle position on axis limits change

		xmin,xmax = event_ax.get_xlim()
		ymin,ymax = event_ax.get_ylim()
		self.circ.center = np.mean([xmin,xmax]), np.mean([ymin,ymax])

	def generate_set_btn(self):
		# Generate the set button for QLineEdit
		
		apply_hbox = QHBoxLayout()
		self.apply_threshold_btn = QPushButton('Set')
		self.apply_threshold_btn.clicked.connect(self.set_threshold_in_parent_le)
		self.apply_threshold_btn.setStyleSheet(self.button_style_sheet)
		apply_hbox.addWidget(QLabel(''),33)
		apply_hbox.addWidget(self.apply_threshold_btn, 33)
		apply_hbox.addWidget(QLabel(''),33)		
		self.canvas.layout.addLayout(apply_hbox)

	def set_threshold_in_parent_le(self):
		# Set the diameter in the parent QLineEdit

		self.parent_le.set_threshold(self.diameter_slider.value())
		self.close()

	def generate_diameter_slider(self):
		# Generate the diameter slider

		self.diameter_slider = QLabeledDoubleSlider()
		diameter_layout = QuickSliderLayout(label='Diameter: ',
										slider=self.diameter_slider,
										slider_initial_value=self.diameter,
										slider_range=self.diameter_slider_range,
										decimal_option=True,
										precision=1.0E-05,
										)
		diameter_layout.setContentsMargins(15,0,15,0)
		self.diameter_slider.valueChanged.connect(self.change_diameter)
		self.canvas.layout.addLayout(diameter_layout)

	def change_diameter(self, value):
		# Change the diameter of the circle
		
		self.diameter = value
		self.circ.set_radius(self.diameter//2)
		self.canvas.canvas.draw_idle()


from PyQt5.QtWidgets import QFrame, QGridLayout, QComboBox, QListWidget, QLabel, QPushButton, QVBoxLayout, QHBoxLayout, QCheckBox, \
	QMessageBox, QWidget
from PyQt5.QtCore import Qt, QSize
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
import gc
from PyQt5.QtGui import QDoubleValidator, QIntValidator

from celldetective.gui.signal_annotator import MeasureAnnotator
from celldetective.gui.signal_annotator2 import SignalAnnotator2
from celldetective.io import get_segmentation_models_list, control_segmentation_napari, get_signal_models_list, \
	control_tracking_btrack, load_experiment_tables, get_pair_signal_models_list
from celldetective.io import locate_segmentation_model, fix_missing_labels, auto_load_number_of_frames, load_frames, locate_signal_model
from celldetective.gui import SegmentationModelLoader, ClassifierWidget, ConfigNeighborhoods, ConfigSegmentationModelTraining, ConfigTracking, SignalAnnotator, ConfigSignalModelTraining, ConfigMeasurements, ConfigSignalAnnotator, TableUI
from celldetective.gui.gui_utils import QHSeperationLine
from celldetective.relative_measurements import rel_measure_at_position
from celldetective.segmentation import segment_at_position, segment_from_threshold_at_position
from celldetective.tracking import track_at_position
from celldetective.measure import measure_at_position
from celldetective.signals import analyze_signals_at_position, analyze_pair_signals_at_position
from celldetective.utils import extract_experiment_channels
import numpy as np
from glob import glob
from natsort import natsorted
import os
import pandas as pd
from celldetective.gui.gui_utils import center_window
from tifffile import imwrite
import json
import psutil
from celldetective.neighborhood import compute_neighborhood_at_position, compute_contact_neighborhood_at_position
from celldetective.gui.gui_utils import FigureCanvas
from celldetective.preprocessing import correct_background_model_free, correct_background_model
from celldetective.utils import _estimate_scale_factor, _extract_channel_indices_from_config, _extract_channel_indices, ConfigSectionMap, _extract_nbr_channels_from_config, _get_img_num_per_channel, normalize_per_channel
from celldetective.gui.gui_utils import ThresholdLineEdit, QuickSliderLayout, help_generic
from celldetective.gui.layouts import CellposeParamsWidget, StarDistParamsWidget, BackgroundModelFreeCorrectionLayout, ProtocolDesignerLayout, BackgroundFitCorrectionLayout
from celldetective.gui import Styles
from celldetective.utils import get_software_location

class ProcessPanel(QFrame, Styles):
	def __init__(self, parent_window, mode):

		super().__init__()
		self.parent_window = parent_window
		self.mode = mode
		self.exp_channels = self.parent_window.exp_channels
		self.exp_dir = self.parent_window.exp_dir
		self.exp_config = self.parent_window.exp_config
		self.movie_prefix = self.parent_window.movie_prefix
		self.threshold_config_targets = None
		self.threshold_config_effectors = None
		self.wells = np.array(self.parent_window.wells,dtype=str)
		self.cellpose_calibrated = False
		self.stardist_calibrated = False

		self.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
		self.grid = QGridLayout(self)
		self.grid.setContentsMargins(5,5,5,5)
		self.generate_header()

	def generate_header(self):

		"""
		Read the mode and prepare a collapsable block to process a specific cell population.

		"""

		panel_title = QLabel(f"PROCESS {self.mode.upper()}   ")
		panel_title.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")

		title_hbox = QHBoxLayout()
		self.grid.addWidget(panel_title, 0, 0, 1, 4, alignment=Qt.AlignCenter)

		self.help_pop_btn = QPushButton()
		self.help_pop_btn.setIcon(icon(MDI6.help_circle,color=self.help_color))
		self.help_pop_btn.setIconSize(QSize(20, 20))
		self.help_pop_btn.clicked.connect(self.help_population)
		self.help_pop_btn.setStyleSheet(self.button_select_all)
		self.help_pop_btn.setToolTip("Help.")
		#self.grid.addWidget(self.help_pop_btn, 0, 0, 1, 3, alignment=Qt.AlignRight)


		self.select_all_btn = QPushButton()
		self.select_all_btn.setIcon(icon(MDI6.checkbox_blank_outline,color="black"))
		self.select_all_btn.setIconSize(QSize(20, 20))
		self.all_ticked = False
		self.select_all_btn.clicked.connect(self.tick_all_actions)
		self.select_all_btn.setStyleSheet(self.button_select_all)
		#self.grid.addWidget(self.select_all_btn, 0, 0, 1, 4, alignment=Qt.AlignLeft)
		#self.to_disable.append(self.all_tc_actions)

		self.collapse_btn = QPushButton()
		self.collapse_btn.setIcon(icon(MDI6.chevron_down,color="black"))
		self.collapse_btn.setIconSize(QSize(25, 25))
		self.collapse_btn.setStyleSheet(self.button_select_all)
		#self.grid.addWidget(self.collapse_btn, 0, 0, 1, 4, alignment=Qt.AlignRight)

		title_hbox.addWidget(self.select_all_btn, 5)
		title_hbox.addWidget(QLabel(), 85, alignment=Qt.AlignCenter)
		title_hbox.addWidget(self.help_pop_btn, 5)
		title_hbox.addWidget(self.collapse_btn, 5)

		self.grid.addLayout(title_hbox, 0,0,1,4)
		self.populate_contents()

		self.grid.addWidget(self.ContentsFrame, 1, 0, 1, 4, alignment=Qt.AlignTop)
		self.collapse_btn.clicked.connect(lambda: self.ContentsFrame.setHidden(not self.ContentsFrame.isHidden()))
		self.collapse_btn.clicked.connect(self.collapse_advanced)
		self.ContentsFrame.hide()

	def collapse_advanced(self):

		effector_open = not self.parent_window.ProcessEffectors.ContentsFrame.isHidden()
		targets_open = not self.parent_window.ProcessTargets.ContentsFrame.isHidden()
		interactions_open = not self.parent_window.NeighPanel.ContentsFrame.isHidden()
		preprocessing_open = not self.parent_window.PreprocessingPanel.ContentsFrame.isHidden()
		is_open = np.array([effector_open, targets_open, interactions_open, preprocessing_open])

		if self.ContentsFrame.isHidden():
			self.collapse_btn.setIcon(icon(MDI6.chevron_down,color="black"))
			self.collapse_btn.setIconSize(QSize(20, 20))
			if len(is_open[is_open])==0:
				self.parent_window.scroll.setMinimumHeight(int(550))
				self.parent_window.adjustSize()
		else:
			self.collapse_btn.setIcon(icon(MDI6.chevron_up,color="black"))
			self.collapse_btn.setIconSize(QSize(20, 20))
			self.parent_window.scroll.setMinimumHeight(min(int(930), int(0.9*self.parent_window.screen_height)))


	def help_population(self):

		"""
		Helper to choose a proper cell population structure.
		"""

		dict_path = os.sep.join([get_software_location(),'celldetective','gui','help','cell-populations.json'])

		with open(dict_path) as f:
			d = json.load(f)

		suggestion = help_generic(d)
		if isinstance(suggestion, str):
			print(f"{suggestion=}")
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Information)
			msgBox.setTextFormat(Qt.RichText)
			msgBox.setText(suggestion)
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None			

	def populate_contents(self):
		self.ContentsFrame = QFrame()
		self.ContentsFrame.setContentsMargins(5,5,5,5)
		self.grid_contents = QGridLayout(self.ContentsFrame)
		self.grid_contents.setContentsMargins(0,0,0,0)
		self.generate_segmentation_options()
		self.generate_tracking_options()
		self.generate_measure_options()
		self.generate_signal_analysis_options()

		self.grid_contents.addWidget(QHSeperationLine(), 9, 0, 1, 4)
		self.view_tab_btn = QPushButton("View table")
		self.view_tab_btn.setStyleSheet(self.button_style_sheet_2)
		self.view_tab_btn.clicked.connect(self.view_table_ui)
		self.view_tab_btn.setToolTip('View table')
		self.view_tab_btn.setIcon(icon(MDI6.table,color="#1565c0"))
		self.view_tab_btn.setIconSize(QSize(20, 20))
		#self.view_tab_btn.setEnabled(False)
		self.grid_contents.addWidget(self.view_tab_btn, 10, 0, 1, 4)

		self.grid_contents.addWidget(QHSeperationLine(), 9, 0, 1, 4)
		self.submit_btn = QPushButton("Submit")
		self.submit_btn.setStyleSheet(self.button_style_sheet_2)
		self.submit_btn.clicked.connect(self.process_population)
		self.grid_contents.addWidget(self.submit_btn, 11, 0, 1, 4)

	def generate_measure_options(self):

		measure_layout = QHBoxLayout()

		self.measure_action = QCheckBox("MEASURE")
		self.measure_action.setStyleSheet("""
			font-size: 10px;
			padding-left: 10px;
			padding-top: 5px;
			""")
		self.measure_action.setIcon(icon(MDI6.eyedropper,color="black"))
		self.measure_action.setIconSize(QSize(20, 20))
		self.measure_action.setToolTip("Measure.")
		measure_layout.addWidget(self.measure_action, 90)
		#self.to_disable.append(self.measure_action_tc)

		self.classify_btn = QPushButton()
		self.classify_btn.setIcon(icon(MDI6.scatter_plot, color="black"))
		self.classify_btn.setIconSize(QSize(20, 20))
		self.classify_btn.setToolTip("Classify data.")
		self.classify_btn.setStyleSheet(self.button_select_all)
		self.classify_btn.clicked.connect(self.open_classifier_ui)
		measure_layout.addWidget(self.classify_btn, 5) #4,2,1,1, alignment=Qt.AlignRight

		self.check_measurements_btn=QPushButton()
		self.check_measurements_btn.setIcon(icon(MDI6.eye_check_outline,color="black"))
		self.check_measurements_btn.setIconSize(QSize(20, 20))
		self.check_measurements_btn.setToolTip("Explore measurements in-situ.")
		self.check_measurements_btn.setStyleSheet(self.button_select_all)
		self.check_measurements_btn.clicked.connect(self.check_measurements)
		measure_layout.addWidget(self.check_measurements_btn, 5)


		self.measurements_config_btn = QPushButton()
		self.measurements_config_btn.setIcon(icon(MDI6.cog_outline,color="black"))
		self.measurements_config_btn.setIconSize(QSize(20, 20))
		self.measurements_config_btn.setToolTip("Configure measurements.")
		self.measurements_config_btn.setStyleSheet(self.button_select_all)
		self.measurements_config_btn.clicked.connect(self.open_measurement_configuration_ui)
		measure_layout.addWidget(self.measurements_config_btn, 5) #4,2,1,1, alignment=Qt.AlignRight

		self.grid_contents.addLayout(measure_layout,5,0,1,4)

	def generate_signal_analysis_options(self):

		signal_layout = QVBoxLayout()
		signal_hlayout = QHBoxLayout()
		self.signal_analysis_action = QCheckBox("SIGNAL ANALYSIS")
		self.signal_analysis_action.setStyleSheet("""
			font-size: 10px;
			padding-left: 10px;
			padding-top: 5px;
			""")
		self.signal_analysis_action.setIcon(icon(MDI6.chart_bell_curve_cumulative,color="black"))
		self.signal_analysis_action.setIconSize(QSize(20, 20))
		self.signal_analysis_action.setToolTip("Detect events in single-cell signals.")
		self.signal_analysis_action.toggled.connect(self.enable_signal_model_list)
		signal_hlayout.addWidget(self.signal_analysis_action, 90)

		self.check_signals_btn = QPushButton()
		self.check_signals_btn.setIcon(icon(MDI6.eye_check_outline,color="black"))
		self.check_signals_btn.setIconSize(QSize(20, 20))
		self.check_signals_btn.clicked.connect(self.check_signals)
		self.check_signals_btn.setToolTip("Explore signals in-situ.")
		self.check_signals_btn.setStyleSheet(self.button_select_all)
		signal_hlayout.addWidget(self.check_signals_btn, 6)

		self.config_signal_annotator_btn = QPushButton()
		self.config_signal_annotator_btn.setIcon(icon(MDI6.cog_outline,color="black"))
		self.config_signal_annotator_btn.setIconSize(QSize(20, 20))
		self.config_signal_annotator_btn.setToolTip("Configure the dynamic visualizer.")
		self.config_signal_annotator_btn.setStyleSheet(self.button_select_all)
		self.config_signal_annotator_btn.clicked.connect(self.open_signal_annotator_configuration_ui)
		signal_hlayout.addWidget(self.config_signal_annotator_btn, 6)

		#self.to_disable.append(self.measure_action_tc)
		signal_layout.addLayout(signal_hlayout)

		signal_model_vbox = QVBoxLayout()
		signal_model_vbox.setContentsMargins(25,0,25,0)

		model_zoo_layout = QHBoxLayout()
		model_zoo_layout.addWidget(QLabel("Model zoo:"),90)

		self.signal_models_list = QComboBox()
		self.signal_models_list.setEnabled(False)
		self.refresh_signal_models()
		#self.to_disable.append(self.cell_models_list)

		self.train_signal_model_btn = QPushButton("TRAIN")
		self.train_signal_model_btn.setToolTip("Open a dialog box to create a new target segmentation model.")
		self.train_signal_model_btn.setIcon(icon(MDI6.redo_variant,color='black'))
		self.train_signal_model_btn.setIconSize(QSize(20, 20))
		self.train_signal_model_btn.setStyleSheet(self.button_style_sheet_3)
		model_zoo_layout.addWidget(self.train_signal_model_btn, 5)
		self.train_signal_model_btn.clicked.connect(self.open_signal_model_config_ui)

		signal_model_vbox.addLayout(model_zoo_layout)
		signal_model_vbox.addWidget(self.signal_models_list)

		signal_layout.addLayout(signal_model_vbox)

		self.grid_contents.addLayout(signal_layout,6,0,1,4)

	def refresh_signal_models(self):
		signal_models = get_signal_models_list()
		self.signal_models_list.clear()
		self.signal_models_list.addItems(signal_models)

	def generate_tracking_options(self):
		grid_track = QHBoxLayout()

		self.track_action = QCheckBox("TRACK")
		self.track_action.setIcon(icon(MDI6.chart_timeline_variant,color="black"))
		self.track_action.setIconSize(QSize(20, 20))
		self.track_action.setToolTip("Track the target cells using bTrack.")
		self.track_action.setStyleSheet("""
			font-size: 10px;
			padding-left: 10px;
			padding-top: 5px;
			""")
		grid_track.addWidget(self.track_action, 75)
		#self.to_disable.append(self.track_action_tc)

		# self.show_track_table_btn = QPushButton()
		# self.show_track_table_btn.setIcon(icon(MDI6.table,color="black"))
		# self.show_track_table_btn.setIconSize(QSize(20, 20))
		# self.show_track_table_btn.setToolTip("Show trajectories table.")
		# self.show_track_table_btn.setStyleSheet(self.button_select_all)
		# #self.show_track_table_btn.clicked.connect(self.display_trajectory_table)
		# self.show_track_table_btn.setEnabled(False)
		# grid_track.addWidget(self.show_track_table_btn, 6)  #4,3,1,1, alignment=Qt.AlignLeft

		self.delete_tracks_btn = QPushButton()
		self.delete_tracks_btn.setIcon(icon(MDI6.trash_can,color="black"))
		self.delete_tracks_btn.setIconSize(QSize(20, 20))
		self.delete_tracks_btn.setToolTip("Delete existing tracks.")
		self.delete_tracks_btn.setStyleSheet(self.button_select_all)
		self.delete_tracks_btn.clicked.connect(self.delete_tracks)
		self.delete_tracks_btn.setEnabled(True)
		self.delete_tracks_btn.hide()
		grid_track.addWidget(self.delete_tracks_btn, 6)  #4,3,1,1, alignment=Qt.AlignLeft

		self.check_tracking_result_btn = QPushButton()
		self.check_tracking_result_btn.setIcon(icon(MDI6.eye_check_outline,color="black"))
		self.check_tracking_result_btn.setIconSize(QSize(20, 20))
		self.check_tracking_result_btn.setToolTip("View raw bTrack output in napari.")
		self.check_tracking_result_btn.setStyleSheet(self.button_select_all)
		self.check_tracking_result_btn.clicked.connect(self.open_napari_tracking)
		self.check_tracking_result_btn.setEnabled(False)
		grid_track.addWidget(self.check_tracking_result_btn, 6)  #4,3,1,1, alignment=Qt.AlignLeft

		self.track_config_btn = QPushButton()
		self.track_config_btn.setIcon(icon(MDI6.cog_outline,color="black"))
		self.track_config_btn.setIconSize(QSize(20, 20))
		self.track_config_btn.setToolTip("Configure tracking.")
		self.track_config_btn.setStyleSheet(self.button_select_all)
		self.track_config_btn.clicked.connect(self.open_tracking_configuration_ui)
		grid_track.addWidget(self.track_config_btn, 6) #4,2,1,1, alignment=Qt.AlignRight

		self.help_track_btn = QPushButton()
		self.help_track_btn.setIcon(icon(MDI6.help_circle,color=self.help_color))
		self.help_track_btn.setIconSize(QSize(20, 20))
		self.help_track_btn.clicked.connect(self.help_tracking)
		self.help_track_btn.setStyleSheet(self.button_select_all)
		self.help_track_btn.setToolTip("Help.")
		grid_track.addWidget(self.help_track_btn, 6) #4,2,1,1, alignment=Qt.AlignRight

		self.grid_contents.addLayout(grid_track, 4, 0, 1,4)

	def delete_tracks(self):

		msgBox = QMessageBox()
		msgBox.setIcon(QMessageBox.Question)
		msgBox.setText("Do you want to erase the tracks? All subsequent annotations will be erased...")
		msgBox.setWindowTitle("Info")
		msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
		returnValue = msgBox.exec()
		if returnValue == QMessageBox.No:
			return None
		elif returnValue == QMessageBox.Yes:
			if os.path.exists(os.sep.join([self.parent_window.pos,'output','tables',f'trajectories_{self.mode}.csv'])):
				os.remove(os.sep.join([self.parent_window.pos,'output','tables',f'trajectories_{self.mode}.csv']))
			if os.path.exists(os.sep.join([self.parent_window.pos,'output','tables',f'trajectories_{self.mode}.pkl'])):
				os.remove(os.sep.join([self.parent_window.pos,'output','tables',f'trajectories_{self.mode}.pkl']))
			if os.path.exists(os.sep.join([self.parent_window.pos,'output','tables',f'napari_{self.mode[:-1]}_trajectories.npy'])):
				os.remove(os.sep.join([self.parent_window.pos,'output','tables',f'napari_{self.mode[:-1]}_trajectories.npy']))
			if os.path.exists(os.sep.join([self.parent_window.pos,'output','tables',f'trajectories_pairs.csv'])):
				os.remove(os.sep.join([self.parent_window.pos,'output','tables',f'trajectories_pairs.csv']))
			self.parent_window.update_position_options()
		else:
			return None

	def generate_segmentation_options(self):

		grid_segment = QHBoxLayout()
		grid_segment.setContentsMargins(0,0,0,0)
		grid_segment.setSpacing(0)

		self.segment_action = QCheckBox("SEGMENT")
		self.segment_action.setStyleSheet("""
			font-size: 10px;
			padding-left: 10px;
			""")
		self.segment_action.setIcon(icon(MDI6.bacteria, color='black'))
		self.segment_action.setToolTip(f"Segment the {self.mode} cells on the images.")
		self.segment_action.toggled.connect(self.enable_segmentation_model_list)
		#self.to_disable.append(self.segment_action)
		grid_segment.addWidget(self.segment_action, 90)

		self.check_seg_btn = QPushButton()
		self.check_seg_btn.setIcon(icon(MDI6.eye_check_outline,color="black"))
		self.check_seg_btn.setIconSize(QSize(20, 20))
		self.check_seg_btn.clicked.connect(self.check_segmentation)
		self.check_seg_btn.setStyleSheet(self.button_select_all)
		self.check_seg_btn.setToolTip("View segmentation output in napari.")
		grid_segment.addWidget(self.check_seg_btn, 5)

		self.help_seg_btn = QPushButton()
		self.help_seg_btn.setIcon(icon(MDI6.help_circle,color=self.help_color))
		self.help_seg_btn.setIconSize(QSize(20, 20))
		self.help_seg_btn.clicked.connect(self.help_segmentation)
		self.help_seg_btn.setStyleSheet(self.button_select_all)
		self.help_seg_btn.setToolTip("Help.")
		grid_segment.addWidget(self.help_seg_btn, 5)
		self.grid_contents.addLayout(grid_segment, 0,0,1,4)

		seg_option_vbox = QVBoxLayout()
		seg_option_vbox.setContentsMargins(25,0,25,0)
		model_zoo_layout = QHBoxLayout()
		model_zoo_layout.addWidget(QLabel("Model zoo:"),90)
		self.seg_model_list = QComboBox()
		#self.to_disable.append(self.tc_seg_model_list)
		self.seg_model_list.setGeometry(50, 50, 200, 30)
		self.init_seg_model_list()


		self.upload_model_btn = QPushButton("UPLOAD")
		self.upload_model_btn.setIcon(icon(MDI6.upload,color="black"))
		self.upload_model_btn.setIconSize(QSize(20, 20))
		self.upload_model_btn.setStyleSheet(self.button_style_sheet_3)
		self.upload_model_btn.setToolTip("Upload a new segmentation model\n(Deep learning or threshold-based).")
		model_zoo_layout.addWidget(self.upload_model_btn, 5)
		self.upload_model_btn.clicked.connect(self.upload_segmentation_model)
		# self.to_disable.append(self.upload_tc_model)

		self.train_btn = QPushButton("TRAIN")
		self.train_btn.setToolTip("Train or retrain a segmentation model\non newly annotated data.")
		self.train_btn.setIcon(icon(MDI6.redo_variant,color='black'))
		self.train_btn.setIconSize(QSize(20, 20))
		self.train_btn.setStyleSheet(self.button_style_sheet_3)
		self.train_btn.clicked.connect(self.open_segmentation_model_config_ui)
		model_zoo_layout.addWidget(self.train_btn, 5)
		# self.train_button_tc.clicked.connect(self.train_stardist_model_tc)
		# self.to_disable.append(self.train_button_tc)

		seg_option_vbox.addLayout(model_zoo_layout)
		seg_option_vbox.addWidget(self.seg_model_list)
		self.seg_model_list.setEnabled(False)
		self.grid_contents.addLayout(seg_option_vbox, 2, 0, 1, 4)

	def help_segmentation(self):

		"""
		Widget with different decision helper decision trees.
		"""

		self.help_w = QWidget()
		self.help_w.setWindowTitle('Helper')
		layout = QVBoxLayout()
		seg_strategy_btn = QPushButton('A guide to choose a segmentation strategy.')
		seg_strategy_btn.setIcon(icon(MDI6.help_circle,color=self.celldetective_blue))
		seg_strategy_btn.setIconSize(QSize(40, 40))
		seg_strategy_btn.setStyleSheet(self.button_style_sheet_5)
		seg_strategy_btn.clicked.connect(self.help_seg_strategy)

		dl_strategy_btn = QPushButton('A guide to choose your Deep learning segmentation strategy.')
		dl_strategy_btn.setIcon(icon(MDI6.help_circle,color=self.celldetective_blue))
		dl_strategy_btn.setIconSize(QSize(40, 40))
		dl_strategy_btn.setStyleSheet(self.button_style_sheet_5)
		dl_strategy_btn.clicked.connect(self.help_seg_dl_strategy)

		layout.addWidget(seg_strategy_btn)
		layout.addWidget(dl_strategy_btn)

		self.help_w.setLayout(layout)
		center_window(self.help_w)
		self.help_w.show()

		return None

	def help_seg_strategy(self):

		"""
		Helper for segmentation strategy between threshold-based and Deep learning.
		"""

		dict_path = os.sep.join([get_software_location(),'celldetective','gui','help','Threshold-vs-DL.json'])

		with open(dict_path) as f:
			d = json.load(f)

		suggestion = help_generic(d)
		if isinstance(suggestion, str):
			print(f"{suggestion=}")
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Information)
			msgBox.setTextFormat(Qt.RichText)
			msgBox.setText(f"The suggested technique is {suggestion}.\nSee a tutorial <a href='https://celldetective.readthedocs.io/en/latest/segment.html'>here</a>.")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

	def help_seg_dl_strategy(self):
		
		"""
		Helper for DL segmentation strategy, between pretrained models and custom models.
		"""

		dict_path = os.sep.join([get_software_location(),'celldetective','gui','help','DL-segmentation-strategy.json'])

		with open(dict_path) as f:
			d = json.load(f)

		suggestion = help_generic(d)
		if isinstance(suggestion, str):
			print(f"{suggestion=}")
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Information)
			msgBox.setText(f"The suggested technique is {suggestion}.")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

	def help_tracking(self):

		"""
		Helper for segmentation strategy between threshold-based and Deep learning.
		"""

		dict_path = os.sep.join([get_software_location(),'celldetective','gui','help','tracking.json'])

		with open(dict_path) as f:
			d = json.load(f)

		suggestion = help_generic(d)
		if isinstance(suggestion, str):
			print(f"{suggestion=}")
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Information)
			msgBox.setTextFormat(Qt.RichText)
			msgBox.setText(f"{suggestion}")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

	def check_segmentation(self):

		if not os.path.exists(os.sep.join([self.parent_window.pos,f'labels_{self.mode}', os.sep])):
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Question)
			msgBox.setText("No labels can be found for this position. Do you want to annotate from scratch?")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.No:
				return None
			else:
				os.mkdir(os.sep.join([self.parent_window.pos,f'labels_{self.mode}']))
				lbl = np.zeros((self.parent_window.shape_x, self.parent_window.shape_y), dtype=int)
				for i in range(self.parent_window.len_movie):
					imwrite(os.sep.join([self.parent_window.pos,f'labels_{self.mode}', str(i).zfill(4)+'.tif']), lbl)

		#self.freeze()
		#QApplication.setOverrideCursor(Qt.WaitCursor)
		test = self.parent_window.locate_selected_position()
		if test:
			print('Memory use: ', dict(psutil.virtual_memory()._asdict()))
			try:
				control_segmentation_napari(self.parent_window.pos, prefix=self.parent_window.movie_prefix, population=self.mode,flush_memory=True)
			except Exception as e:
				msgBox = QMessageBox()
				msgBox.setIcon(QMessageBox.Warning)
				msgBox.setText(str(e))
				msgBox.setWindowTitle("Warning")
				msgBox.setStandardButtons(QMessageBox.Ok)
				returnValue = msgBox.exec()

				msgBox = QMessageBox()
				msgBox.setIcon(QMessageBox.Question)
				msgBox.setText("Would you like to pass empty frames to fix the asymmetry?")
				msgBox.setWindowTitle("Question")
				msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
				returnValue = msgBox.exec()
				if returnValue == QMessageBox.Yes:
					print('Fixing the missing labels...')
					fix_missing_labels(self.parent_window.pos, prefix=self.parent_window.movie_prefix,population=self.mode)
					try:
						control_segmentation_napari(self.parent_window.pos, prefix=self.parent_window.movie_prefix, population=self.mode,flush_memory=True)
					except Exception as e:
						print(f'Error {e}')
						return None
				else:
					return None

			gc.collect()

	def check_signals(self):

		test = self.parent_window.locate_selected_position()
		if test:
			self.SignalAnnotator = SignalAnnotator(self)
			self.SignalAnnotator.show()

	def check_measurements(self):

		test = self.parent_window.locate_selected_position()
		if test:
			self.MeasureAnnotator = MeasureAnnotator(self)
			self.MeasureAnnotator.show()

	def enable_segmentation_model_list(self):
		if self.segment_action.isChecked():
			self.seg_model_list.setEnabled(True)
		else:
			self.seg_model_list.setEnabled(False)

	def enable_signal_model_list(self):
		if self.signal_analysis_action.isChecked():
			self.signal_models_list.setEnabled(True)
		else:
			self.signal_models_list.setEnabled(False)

	def init_seg_model_list(self):

		self.seg_model_list.clear()
		self.seg_models = get_segmentation_models_list(mode=self.mode, return_path=False)
		thresh = 40
		self.models_truncated = [m[:thresh - 3]+'...' if len(m)>thresh else m for m in self.seg_models]
		#self.seg_model_list.addItems(models_truncated)

		self.seg_models_generic = get_segmentation_models_list(mode="generic", return_path=False)
		self.seg_models.append('Threshold')
		self.seg_models.extend(self.seg_models_generic)

		#self.seg_models_generic.insert(0,'Threshold')
		self.seg_model_list.addItems(self.seg_models)

		for i in range(len(self.seg_models)):
			self.seg_model_list.setItemData(i, self.seg_models[i], Qt.ToolTipRole)

		self.seg_model_list.insertSeparator(len(self.models_truncated))


		#if ("live_nuclei_channel" in self.exp_channels)*("dead_nuclei_channel" in self.exp_channels):
		# 	print("both channels found")
		# 	index = self.tc_seg_model_list.findText("MCF7_Hoescht_PI_w_primary_NK", Qt.MatchFixedString)
		# 	if index >= 0:
		# 		self.tc_seg_model_list.setCurrentIndex(index)
		# elif ("live_nuclei_channel" in self.exp_channels)*("dead_nuclei_channel" not in self.exp_channels):
		# 	index = self.tc_seg_model_list.findText("MCF7_Hoescht_w_primary_NK", Qt.MatchFixedString)
		# 	if index >= 0:
		# 		self.tc_seg_model_list.setCurrentIndex(index)
		# elif ("live_nuclei_channel" not in self.exp_channels)*("dead_nuclei_channel" in self.exp_channels):
		# 	index = self.tc_seg_model_list.findText("MCF7_PI_w_primary_NK", Qt.MatchFixedString)
		# 	if index >= 0:
		# 		self.tc_seg_model_list.setCurrentIndex(index)
		# elif ("live_nuclei_channel" not in self.exp_channels)*("dead_nuclei_channel" not in self.exp_channels)*("adhesion_channel" in self.exp_channels):
		# 	index = self.tc_seg_model_list.findText("RICM", Qt.MatchFixedString)
		# 	if index >= 0:
		# 		self.tc_seg_model_list.setCurrentIndex(index)

	def tick_all_actions(self):
		self.switch_all_ticks_option()
		if self.all_ticked:
			self.select_all_btn.setIcon(icon(MDI6.checkbox_outline,color="black"))
			self.select_all_btn.setIconSize(QSize(20, 20))
			self.segment_action.setChecked(True)
		else:
			self.select_all_btn.setIcon(icon(MDI6.checkbox_blank_outline,color="black"))
			self.select_all_btn.setIconSize(QSize(20, 20))
			self.segment_action.setChecked(False)

	def switch_all_ticks_option(self):
		if self.all_ticked == True:
			self.all_ticked = False
		else:
			self.all_ticked = True

	def upload_segmentation_model(self):
		print('Load a segmentation model or pipeline...')
		self.SegModelLoader = SegmentationModelLoader(self)
		self.SegModelLoader.show()

	def open_tracking_configuration_ui(self):
		print('Set the tracking parameters...')
		self.ConfigTracking = ConfigTracking(self)
		self.ConfigTracking.show()

	def open_signal_model_config_ui(self):
		print('Set the training parameters for new signal models...')
		self.ConfigSignalTrain = ConfigSignalModelTraining(self)
		self.ConfigSignalTrain.show()

	def open_segmentation_model_config_ui(self):
		print('Set the training parameters for a new segmentation model...')
		self.ConfigSegmentationTrain = ConfigSegmentationModelTraining(self)
		self.ConfigSegmentationTrain.show()

	def open_measurement_configuration_ui(self):
		print('Set the measurements to be performed...')
		self.ConfigMeasurements = ConfigMeasurements(self)
		self.ConfigMeasurements.show()

	def open_classifier_ui(self):

		self.load_available_tables()
		if self.df is None:

			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("No table was found...")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None
			else:
				return None
		else:
			self.ClassifierWidget = ClassifierWidget(self)
			self.ClassifierWidget.show()

	def open_signal_annotator_configuration_ui(self):
		self.ConfigSignalAnnotator = ConfigSignalAnnotator(self)
		self.ConfigSignalAnnotator.show()

	def process_population(self):

		if self.parent_window.well_list.currentText()=="*":
			self.well_index = np.linspace(0,len(self.wells)-1,len(self.wells),dtype=int)
		else:
			self.well_index = [self.parent_window.well_list.currentIndex()]
			print(f"Processing well {self.parent_window.well_list.currentText()}...")

		# self.freeze()
		# QApplication.setOverrideCursor(Qt.WaitCursor)

		if self.mode=="targets":
			self.threshold_config = self.threshold_config_targets
		elif self.mode=="effectors":
			self.threshold_config = self.threshold_config_effectors

		self.load_available_tables()

		if self.df is not None and self.segment_action.isChecked():
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Question)
			msgBox.setText("Measurement tables have been found... Re-segmenting may create mismatches between the cell labels and the associated measurements. Do you want to erase the tables post-segmentation?")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.No:
				pass
			elif returnValue == QMessageBox.Cancel:
				return None
			else:
				print('erase tabs!')
				tabs = [pos+os.sep.join(['output', 'tables', f'trajectories_{self.mode}.csv']) for pos in self.df_pos_info['pos_path'].unique()]
				for t in tabs:
					if os.path.exists(t.replace('.csv','.pkl')):
						os.remove(t.replace('.csv','.pkl'))
					os.remove(t)				

		loop_iter=0

		if self.parent_window.position_list.currentText()=="*":
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Question)
			msgBox.setText("If you continue, all positions will be processed.\nDo you want to proceed?")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.No:
				return None

		if self.seg_model_list.currentIndex() > len(self.models_truncated):
			self.model_name = self.seg_models[self.seg_model_list.currentIndex()-1]
		else:
			self.model_name = self.seg_models[self.seg_model_list.currentIndex()]
		print(self.model_name, self.seg_model_list.currentIndex())

		if self.model_name.startswith('CP') and self.model_name in self.seg_models_generic and not self.cellpose_calibrated:

			self.diamWidget = CellposeParamsWidget(self, model_name=self.model_name)
			self.diamWidget.show()
			return None

		if self.model_name.startswith('SD') and self.model_name in self.seg_models_generic and not self.stardist_calibrated:

			self.diamWidget = StarDistParamsWidget(self, model_name = self.model_name)
			self.diamWidget.show()
			return None


		for w_idx in self.well_index:

			pos = self.parent_window.positions[w_idx]
			if self.parent_window.position_list.currentText()=="*":
				pos_indices = np.linspace(0,len(pos)-1,len(pos),dtype=int)
				print("Processing all positions...")
			else:
				pos_indices = natsorted([pos.index(self.parent_window.position_list.currentText())])
				print(f"Processing position {self.parent_window.position_list.currentText()}...")

			well = self.parent_window.wells[w_idx]

			for pos_idx in pos_indices:

				self.pos = natsorted(glob(well+f"{os.path.split(well)[-1].replace('W','').replace(os.sep,'')}*/"))[pos_idx]
				print(f"Position {self.pos}...\nLoading stack movie...")

				if not os.path.exists(self.pos + 'output/'):
					os.mkdir(self.pos + 'output/')
				if not os.path.exists(self.pos + 'output/tables/'):
					os.mkdir(self.pos + 'output/tables/')

				if self.segment_action.isChecked():

					if len(glob(os.sep.join([self.pos, f'labels_{self.mode}','*.tif'])))>0 and self.parent_window.position_list.currentText()!="*":
						msgBox = QMessageBox()
						msgBox.setIcon(QMessageBox.Question)
						msgBox.setText("Labels have already been produced for this position. Do you want to segment again?")
						msgBox.setWindowTitle("Info")
						msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
						returnValue = msgBox.exec()
						if returnValue == QMessageBox.No:
							return None

					if (self.seg_model_list.currentText()=="Threshold"):
						if self.threshold_config is None:
							msgBox = QMessageBox()
							msgBox.setIcon(QMessageBox.Warning)
							msgBox.setText("Please set a threshold configuration from the upload menu first. Abort.")
							msgBox.setWindowTitle("Warning")
							msgBox.setStandardButtons(QMessageBox.Ok)
							returnValue = msgBox.exec()
							if returnValue == QMessageBox.Ok:
								return None
						else:
							print(f"Segmentation from threshold config: {self.threshold_config}")
							segment_from_threshold_at_position(self.pos, self.mode, self.threshold_config, threads=self.parent_window.parent_window.n_threads)
					else:
						segment_at_position(self.pos, self.mode, self.model_name, stack_prefix=self.parent_window.movie_prefix, use_gpu=self.parent_window.parent_window.use_gpu, threads=self.parent_window.parent_window.n_threads)

				if self.track_action.isChecked():
					if os.path.exists(os.sep.join([self.pos, 'output', 'tables', f'trajectories_{self.mode}.csv'])) and self.parent_window.position_list.currentText()!="*":
						msgBox = QMessageBox()
						msgBox.setIcon(QMessageBox.Question)
						msgBox.setText("A trajectory set already exists. Previously annotated data for\nthis position will be lost. Do you want to proceed?")
						msgBox.setWindowTitle("Info")
						msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
						returnValue = msgBox.exec()
						if returnValue == QMessageBox.No:
							return None
					track_at_position(self.pos, self.mode, threads=self.parent_window.parent_window.n_threads)

				if self.measure_action.isChecked():
					measure_at_position(self.pos, self.mode, threads=self.parent_window.parent_window.n_threads)

				table = os.sep.join([self.pos, 'output', 'tables', f'trajectories_{self.mode}.csv'])
				if self.signal_analysis_action.isChecked() and os.path.exists(table):
					print('table exists')
					table = pd.read_csv(table)
					cols = list(table.columns)
					print(table, cols)
					if 'class_color' in cols:
						print(cols, 'class_color in cols')
						colors = list(table['class_color'].to_numpy())
						if 'tab:orange' in colors or 'tab:cyan' in colors:
							if self.parent_window.position_list.currentText()!="*":
								msgBox = QMessageBox()
								msgBox.setIcon(QMessageBox.Question)
								msgBox.setText("The signals of the cells in the position appear to have been annotated... Do you want to proceed?")
								msgBox.setWindowTitle("Info")
								msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
								returnValue = msgBox.exec()
								if returnValue == QMessageBox.No:
									return None
					analyze_signals_at_position(self.pos, self.signal_models_list.currentText(), self.mode)


			# self.stack = None
		self.parent_window.update_position_options()
		if self.segment_action.isChecked():
			self.segment_action.setChecked(False)

		# QApplication.restoreOverrideCursor()
		# self.unfreeze()

	# def view_current_stack_with_scale_bar(self):

	# 	self.parent_window.locate_image()
	# 	if self.parent_window.current_stack is not None:
	# 		self.viewer = CellSizeViewer(
	# 									  initial_diameter = float(self.diameter_le.text().replace(',', '.')),
	# 			parent_le = self.diameter_le,
	# 									  stack_path=self.parent_window.current_stack,
	# 									  window_title=f'Position {self.parent_window.position_list.currentText()}',
	# 									  frame_slider = True,
	# 									  contrast_slider = True,
	# 									  channel_cb = True,
	# 									  channel_names = self.parent_window.exp_channels,
	# 									  n_channels = self.parent_window.nbr_channels,
	# 									  PxToUm = 1,
	# 									 )
	# 		self.viewer.show()



	def open_napari_tracking(self):
		print(f'View the tracks before post-processing for position {self.parent_window.pos} in napari...')
		control_tracking_btrack(self.parent_window.pos, prefix=self.parent_window.movie_prefix, population=self.mode, threads=self.parent_window.parent_window.n_threads)

	def view_table_ui(self):

		print('Load table...')
		self.load_available_tables()

		if self.df is not None:
			plot_mode = 'plot_track_signals'
			if 'TRACK_ID' not in list(self.df.columns):
				plot_mode = 'static'
			self.tab_ui = TableUI(self.df, f"Well {self.parent_window.well_list.currentText()}; Position {self.parent_window.position_list.currentText()}", population=self.mode, plot_mode=plot_mode)
			self.tab_ui.show()
		else:
			print('Table could not be loaded...')
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("No table could be loaded...")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

	# def interpret_pos_location(self):

	# 	"""
	# 	Read the well/position selection from the control panel to decide which data to load
	# 	Set position_indices to None if all positions must be taken

	# 	"""

	# 	if self.well_option==len(self.wells):
	# 		self.well_indices = np.arange(len(self.wells))
	# 	else:
	# 		self.well_indices = np.array([self.well_option],dtype=int)

	# 	if self.position_option==0:
	# 		self.position_indices = None
	# 	else:
	# 		self.position_indices = np.array([self.position_option],dtype=int)

	def load_available_tables(self):

		"""
		Load the tables of the selected wells/positions from the control Panel for the population of interest

		"""

		self.well_option = self.parent_window.well_list.currentIndex()
		if self.well_option==len(self.wells):
			wo = '*'
		else:
			wo = self.well_option
		self.position_option = self.parent_window.position_list.currentIndex()
		if self.position_option==0:
			po = '*'
		else:
			po = self.position_option - 1

		self.df, self.df_pos_info = load_experiment_tables(self.exp_dir, well_option=wo, position_option=po, population=self.mode, return_pos_info=True)
		if self.df is None:
			print('No table could be found...')

	def set_cellpose_scale(self):

		scale = self.parent_window.PxToUm * float(self.diamWidget.diameter_le.get_threshold()) / 30.0
		if self.model_name=="CP_nuclei":
			scale = self.parent_window.PxToUm * float(self.diamWidget.diameter_le.get_threshold()) / 17.0
		flow_thresh = self.diamWidget.flow_slider.value()
		cellprob_thresh = self.diamWidget.cellprob_slider.value()
		model_complete_path = locate_segmentation_model(self.model_name)
		input_config_path = model_complete_path+"config_input.json"
		new_channels = [self.diamWidget.cellpose_channel_cb[i].currentText() for i in range(2)]
		print(new_channels)
		with open(input_config_path) as config_file:
			input_config = json.load(config_file)

		input_config['spatial_calibration'] = scale
		input_config['channels'] = new_channels
		input_config['flow_threshold'] = flow_thresh
		input_config['cellprob_threshold'] = cellprob_thresh
		with open(input_config_path, 'w') as f:
			json.dump(input_config, f, indent=4)

		self.cellpose_calibrated = True
		print('model scale automatically computed: ', scale)
		self.diamWidget.close()
		self.process_population()

	def set_stardist_scale(self):

		model_complete_path = locate_segmentation_model(self.model_name)
		input_config_path = model_complete_path+"config_input.json"
		new_channels = [self.diamWidget.stardist_channel_cb[i].currentText() for i in range(len(self.diamWidget.stardist_channel_cb))]
		with open(input_config_path) as config_file:
			input_config = json.load(config_file)

		input_config['channels'] = new_channels
		with open(input_config_path, 'w') as f:
			json.dump(input_config, f, indent=4)

		self.stardist_calibrated = True
		self.diamWidget.close()
		self.process_population()


class NeighPanel(QFrame, Styles):
	def __init__(self, parent_window):

		super().__init__()
		self.parent_window = parent_window
		self.exp_channels = self.parent_window.exp_channels
		self.exp_dir = self.parent_window.exp_dir
		self.wells = np.array(self.parent_window.wells,dtype=str)
		self.protocols = []
		self.mode='neighborhood'

		self.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
		self.grid = QGridLayout(self)
		self.generate_header()

	def generate_header(self):

		"""
		Read the mode and prepare a collapsable block to process a specific cell population.

		"""

		panel_title = QLabel(f"INTERACTIONS")
		panel_title.setStyleSheet(self.block_title)

		self.grid.addWidget(panel_title, 0, 0, 1, 4, alignment=Qt.AlignCenter)

		# self.select_all_btn = QPushButton()
		# self.select_all_btn.setIcon(icon(MDI6.checkbox_blank_outline,color="black"))
		# self.select_all_btn.setIconSize(QSize(20, 20))
		# self.all_ticked = False
		# self.select_all_btn.setStyleSheet(self.button_select_all)
		# self.grid.addWidget(self.select_all_btn, 0, 0, 1, 4, alignment=Qt.AlignLeft)

		self.collapse_btn = QPushButton()
		self.collapse_btn.setIcon(icon(MDI6.chevron_down,color="black"))
		self.collapse_btn.setIconSize(QSize(25, 25))
		self.collapse_btn.setStyleSheet(self.button_select_all)
		self.grid.addWidget(self.collapse_btn, 0, 0, 1, 4, alignment=Qt.AlignRight)

		self.populate_contents()

		self.grid.addWidget(self.ContentsFrame, 1, 0, 1, 4, alignment=Qt.AlignTop)
		self.collapse_btn.clicked.connect(lambda: self.ContentsFrame.setHidden(not self.ContentsFrame.isHidden()))
		self.collapse_btn.clicked.connect(self.collapse_advanced)
		self.ContentsFrame.hide()

	def collapse_advanced(self):

		effector_open = not self.parent_window.ProcessEffectors.ContentsFrame.isHidden()
		targets_open = not self.parent_window.ProcessTargets.ContentsFrame.isHidden()
		interactions_open = not self.parent_window.NeighPanel.ContentsFrame.isHidden()
		preprocessing_open = not self.parent_window.PreprocessingPanel.ContentsFrame.isHidden()
		is_open = np.array([effector_open, targets_open, interactions_open, preprocessing_open])

		if self.ContentsFrame.isHidden():
			self.collapse_btn.setIcon(icon(MDI6.chevron_down,color="black"))
			self.collapse_btn.setIconSize(QSize(20, 20))
			if len(is_open[is_open])==0:
				self.parent_window.scroll.setMinimumHeight(int(550))
				self.parent_window.adjustSize()
		else:
			self.collapse_btn.setIcon(icon(MDI6.chevron_up,color="black"))
			self.collapse_btn.setIconSize(QSize(20, 20))
			self.parent_window.scroll.setMinimumHeight(min(int(1000), int(0.9*self.parent_window.screen_height)))


	def populate_contents(self):

		self.ContentsFrame = QFrame()
		self.grid_contents = QGridLayout(self.ContentsFrame)
		self.grid_contents.setContentsMargins(0,0,0,0)
		self.grid_contents.setSpacing(3)

		# Button to compute the neighborhoods
		neigh_option_hbox = QHBoxLayout()
		self.neigh_action = QCheckBox('NEIGHBORHOODS')
		self.neigh_action.setStyleSheet("""
					font-size: 10px;
					padding-left: 10px;
					padding-top: 5px;
					""")
		#self.neigh_action.setIcon(icon(MDI6.eyedropper, color="black"))
		#self.neigh_action.setIconSize(QSize(20, 20))
		self.neigh_action.setToolTip(
			"Compute neighborhoods in list below.")

		neigh_option_hbox.addWidget(self.neigh_action,90)


		self.help_neigh_btn = QPushButton()
		self.help_neigh_btn.setIcon(icon(MDI6.help_circle,color=self.help_color))
		self.help_neigh_btn.setIconSize(QSize(20, 20))
		self.help_neigh_btn.clicked.connect(self.help_neighborhood)
		self.help_neigh_btn.setStyleSheet(self.button_select_all)
		self.help_neigh_btn.setToolTip("Help.")
		neigh_option_hbox.addWidget(self.help_neigh_btn,5,alignment=Qt.AlignRight)

		self.grid_contents.addLayout(neigh_option_hbox, 1,0,1,4)


		neigh_options_layout = QVBoxLayout()

		neigh_options_vbox = QVBoxLayout()

		# DISTANCE NEIGHBORHOOD
		dist_neigh_hbox = QHBoxLayout()
		dist_neigh_hbox.setContentsMargins(0,0,0,0)
		dist_neigh_hbox.setSpacing(0)

		self.dist_neigh_action = QLabel("ISOTROPIC DISTANCE THRESHOLD")
		self.dist_neigh_action.setStyleSheet(self.action_lbl_style_sheet)
		#self.dist_neigh_action.setIcon(icon(MDI6.circle_expand, color='black'))
		self.dist_neigh_action.setToolTip("")
		self.dist_neigh_action.setToolTip("Define an isotropic neighborhood between the center of mass\nof the cells, within a threshold distance.")
		#self.segment_action.toggled.connect(self.enable_segmentation_model_list)
		#self.to_disable.append(self.segment_action)

		self.config_distance_neigh_btn = QPushButton()
		self.config_distance_neigh_btn.setIcon(icon(MDI6.plus,color="black"))
		self.config_distance_neigh_btn.setIconSize(QSize(20, 20))
		self.config_distance_neigh_btn.setToolTip("Configure.")
		self.config_distance_neigh_btn.setStyleSheet(self.button_select_all)
		self.config_distance_neigh_btn.clicked.connect(self.open_config_distance_threshold_neighborhood)
		dist_neigh_hbox.addWidget(self.config_distance_neigh_btn,5)
		dist_neigh_hbox.addWidget(self.dist_neigh_action, 95)
		neigh_options_vbox.addLayout(dist_neigh_hbox)

		# CONTACT NEIGHBORHOOD
		contact_neighborhood_layout = QHBoxLayout()
		contact_neighborhood_layout.setContentsMargins(0,0,0,0)
		contact_neighborhood_layout.setSpacing(0)

		self.contact_neigh_action = QLabel("MASK CONTACT")
		self.contact_neigh_action.setToolTip("Identify touching cell masks, within a threshold edge distance.")
		self.contact_neigh_action.setStyleSheet(self.action_lbl_style_sheet)
		#self.contact_neigh_action.setIcon(icon(MDI6.transition_masked, color='black'))
		self.contact_neigh_action.setToolTip("")

		self.config_contact_neigh_btn = QPushButton()
		self.config_contact_neigh_btn.setIcon(icon(MDI6.plus,color="black"))
		self.config_contact_neigh_btn.setIconSize(QSize(20, 20))
		self.config_contact_neigh_btn.setToolTip("Configure.")
		self.config_contact_neigh_btn.setStyleSheet(self.button_select_all)
		self.config_contact_neigh_btn.clicked.connect(self.open_config_contact_neighborhood)
		contact_neighborhood_layout.addWidget(self.config_contact_neigh_btn,5)
		contact_neighborhood_layout.addWidget(self.contact_neigh_action, 95)
		neigh_options_vbox.addLayout(contact_neighborhood_layout)
		#self.grid_contents.addLayout(neigh_options_vbox, 2,0,1,4)

		#self.grid_contents.addWidget(QHSeperationLine(), 3, 0, 1, 4)

		self.delete_protocol_btn = QPushButton('')
		self.delete_protocol_btn.setStyleSheet(self.button_select_all)
		self.delete_protocol_btn.setIcon(icon(MDI6.trash_can, color="black"))
		self.delete_protocol_btn.setToolTip("Remove a neighborhood computation.")
		self.delete_protocol_btn.setIconSize(QSize(20, 20))
		self.delete_protocol_btn.clicked.connect(self.remove_protocol_from_list)

		self.protocol_list_lbl = QLabel('Neighborhoods to compute: ')
		self.protocol_list = QListWidget()
		self.protocol_list.setToolTip("Neighborhoods to compute sequentially.")

		list_header_layout = QHBoxLayout()
		list_header_layout.addWidget(self.protocol_list_lbl)
		list_header_layout.addWidget(self.delete_protocol_btn, alignment=Qt.AlignRight)
		#self.grid_contents.addLayout(list_header_layout, 4, 0, 1, 4)
		#self.grid_contents.addWidget(self.protocol_list, 5, 0, 1, 4)

		neigh_options_layout.addLayout(neigh_options_vbox)
		neigh_options_layout.addWidget(QHSeperationLine())
		neigh_options_layout.addLayout(list_header_layout)
		neigh_options_layout.addWidget(self.protocol_list)

		neigh_options_layout.setContentsMargins(30,5,30,5)
		neigh_options_layout.setSpacing(1)
		self.grid_contents.addLayout(neigh_options_layout, 5, 0, 1, 4)


		rel_layout = QHBoxLayout()
		self.measure_pairs_action = QCheckBox("MEASURE PAIRS")
		self.measure_pairs_action.setStyleSheet("""
					font-size: 10px;
					padding-left: 10px;
					padding-top: 5px;
					""")
		self.measure_pairs_action.setIcon(icon(MDI6.eyedropper, color="black"))
		self.measure_pairs_action.setIconSize(QSize(20, 20))
		self.measure_pairs_action.setToolTip("Measure the relative quantities defined for the cell pairs, for all neighborhoods.")
		rel_layout.addWidget(self.measure_pairs_action, 90)

		# self.visu_btn = QPushButton()
		# self.visu_btn.setIcon(icon(MDI6.eye_check_outline,color="black"))
		# self.visu_btn.setIconSize(QSize(20, 20))
		# self.visu_btn.clicked.connect(self.check_measurements2)
		# self.visu_btn.setToolTip("Open measurement annotator for two populations.")
		# self.visu_btn.setStyleSheet(self.button_select_all)
		# self.grid_contents.addWidget(self.visu_btn, 1,1,1,1,alignment=Qt.AlignRight)
		# rel_layout.addWidget(self.visu_btn, 6)

		# self.config_rel_annotator_btn = QPushButton()
		# self.config_rel_annotator_btn.setIcon(icon(MDI6.cog_outline, color="black"))
		# self.config_rel_annotator_btn.setIconSize(QSize(20, 20))
		# self.config_rel_annotator_btn.setToolTip("Configure the animation of the annotation tool.")
		# self.config_rel_annotator_btn.setStyleSheet(self.button_select_all)
		# self.config_rel_annotator_btn.clicked.connect(self.open_signal_annotator_configuration_ui)
		# # self.grid_contents.addWidget(self.config_rel_annotator_btn, 1,2,1,1, alignment=Qt.AlignRight)
		# rel_layout.addWidget(self.config_rel_annotator_btn, 6)
		self.grid_contents.addLayout(rel_layout, 6, 0, 1, 4)

		signal_layout = QVBoxLayout()
		signal_hlayout = QHBoxLayout()
		self.signal_analysis_action = QCheckBox("PAIR SIGNAL ANALYSIS")
		self.signal_analysis_action.setStyleSheet("""
		font-size: 10px;
		padding-left: 10px;
		padding-top: 5px;""")

		self.signal_analysis_action.setIcon(icon(MDI6.chart_bell_curve_cumulative, color="black"))
		self.signal_analysis_action.setIconSize(QSize(20, 20))
		self.signal_analysis_action.setToolTip("Detect cell pair events using a DL model.")
		self.signal_analysis_action.toggled.connect(self.enable_signal_model_list)
		signal_hlayout.addWidget(self.signal_analysis_action, 90)

		self.check_signals_btn = QPushButton()
		self.check_signals_btn.setIcon(icon(MDI6.eye_check_outline, color="black"))
		self.check_signals_btn.setIconSize(QSize(20, 20))
		self.check_signals_btn.clicked.connect(self.check_signals2)
		self.check_signals_btn.setToolTip("Annotate dynamic cell pairs.")
		self.check_signals_btn.setStyleSheet(self.button_select_all)
		signal_hlayout.addWidget(self.check_signals_btn, 6)

		self.config_signal_annotator_btn = QPushButton()
		self.config_signal_annotator_btn.setIcon(icon(MDI6.cog_outline, color="black"))
		self.config_signal_annotator_btn.setIconSize(QSize(20, 20))
		self.config_signal_annotator_btn.setToolTip("Configure the animation of the annotation tool.")
		self.config_signal_annotator_btn.setStyleSheet(self.button_select_all)
		self.config_signal_annotator_btn.clicked.connect(self.open_signal_annotator_configuration_ui)
		signal_hlayout.addWidget(self.config_signal_annotator_btn, 6)
		signal_layout.addLayout(signal_hlayout)
		# self.to_disable.append(self.measure_action_tc)
		pair_signal_model_vbox = QVBoxLayout()
		pair_signal_model_vbox.setContentsMargins(25, 0, 25, 0)

		pair_model_zoo_layout = QHBoxLayout()
		pair_model_zoo_layout.addWidget(QLabel("Model zoo:"), 90)

		self.pair_signal_models_list = QComboBox()
		self.pair_signal_models_list.setEnabled(False)
		self.refresh_signal_models()
		# self.to_disable.append(self.cell_models_list)

		self.pair_train_signal_model_btn = QPushButton("TRAIN")
		self.pair_train_signal_model_btn.setToolTip("Train a cell pair event detection model.")
		self.pair_train_signal_model_btn.setIcon(icon(MDI6.redo_variant, color='black'))
		self.pair_train_signal_model_btn.setIconSize(QSize(20, 20))
		self.pair_train_signal_model_btn.setStyleSheet(self.button_style_sheet_3)
		pair_model_zoo_layout.addWidget(self.pair_train_signal_model_btn, 5)
		self.pair_train_signal_model_btn.clicked.connect(self.open_signal_model_config_ui)

		pair_signal_model_vbox.addLayout(pair_model_zoo_layout)
		pair_signal_model_vbox.addWidget(self.pair_signal_models_list)

		signal_layout.addLayout(pair_signal_model_vbox)
		self.grid_contents.addLayout(signal_layout, 7, 0, 1, 4)
		self.grid_contents.addWidget(QHSeperationLine(), 11, 0, 1, 4)

		self.view_tab_btn = QPushButton("View table")
		self.view_tab_btn.setStyleSheet(self.button_style_sheet_2)
		self.view_tab_btn.clicked.connect(self.view_table_ui)
		self.view_tab_btn.setToolTip('View table')
		self.view_tab_btn.setIcon(icon(MDI6.table,color="#1565c0"))
		self.view_tab_btn.setIconSize(QSize(20, 20))
		#self.view_tab_btn.setEnabled(False)
		self.grid_contents.addWidget(self.view_tab_btn, 12, 0, 1, 4)

		#self.grid_contents.addWidget(QLabel(''), 12, 0, 1, 4)

		self.submit_btn = QPushButton("Submit")
		self.submit_btn.setStyleSheet(self.button_style_sheet_2)
		self.submit_btn.setToolTip("Compute the neighborhoods of the selected positions.")
		self.submit_btn.clicked.connect(self.process_neighborhood)
		self.grid_contents.addWidget(self.submit_btn, 14, 0, 1, 4)

		self.neigh_action.toggled.connect(self.activate_neigh_options)
		self.neigh_action.setChecked(True)
		self.neigh_action.setChecked(False)


	def help_neighborhood(self):

		"""
		Helper for neighborhood strategy.
		"""

		dict_path = os.sep.join([get_software_location(),'celldetective','gui','help','neighborhood.json'])

		with open(dict_path) as f:
			d = json.load(f)

		suggestion = help_generic(d)
		if isinstance(suggestion, str):
			print(f"{suggestion=}")
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Information)
			msgBox.setTextFormat(Qt.RichText)
			msgBox.setText(f"{suggestion}\nSee a tutorial <a href='https://celldetective.readthedocs.io/en/latest/interactions.html#neighborhood'>here</a>.")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None


	def load_available_tables(self):

		"""
		Load the tables of the selected wells/positions from the control Panel for the population of interest

		"""

		self.well_option = self.parent_window.well_list.currentIndex()
		if self.well_option==len(self.wells):
			wo = '*'
		else:
			wo = self.well_option
		self.position_option = self.parent_window.position_list.currentIndex()
		if self.position_option==0:
			po = '*'
		else:
			po = self.position_option - 1

		self.df, self.df_pos_info = load_experiment_tables(self.exp_dir, well_option=wo, position_option=po, population="pairs", return_pos_info=True)
		if self.df is None:
			print('No table could be found...')


	def view_table_ui(self):

		print('Load table...')
		self.load_available_tables()

		if self.df is not None:
			plot_mode = 'static'
			self.tab_ui = TableUI(self.df, f"Well {self.parent_window.well_list.currentText()}; Position {self.parent_window.position_list.currentText()}", population='pairs', plot_mode=plot_mode)
			self.tab_ui.show()
		else:
			print('Table could not be loaded...')
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("No table could be loaded...")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None


	def activate_neigh_options(self):

		if self.neigh_action.isChecked():
			self.dist_neigh_action.setEnabled(True)
			self.contact_neigh_action.setEnabled(True)
			self.config_distance_neigh_btn.setEnabled(True)
			self.config_contact_neigh_btn.setEnabled(True)
			self.protocol_list_lbl.setEnabled(True)
			self.protocol_list.setEnabled(True)
			self.delete_protocol_btn.setEnabled(True)
		else:
			self.dist_neigh_action.setEnabled(False)
			self.contact_neigh_action.setEnabled(False)
			self.config_distance_neigh_btn.setEnabled(False)
			self.config_contact_neigh_btn.setEnabled(False)
			self.protocol_list_lbl.setEnabled(False)
			self.protocol_list.setEnabled(False)
			self.delete_protocol_btn.setEnabled(False)

	def refresh_signal_models(self):
		signal_models = get_pair_signal_models_list()
		self.pair_signal_models_list.clear()
		self.pair_signal_models_list.addItems(signal_models)

	def open_signal_annotator_configuration_ui(self):
		self.ConfigSignalAnnotator = ConfigSignalAnnotator(self)
		self.ConfigSignalAnnotator.show()

	def open_signal_model_config_ui(self):
		self.ConfigSignalTrain = ConfigSignalModelTraining(self, signal_mode='pairs')
		self.ConfigSignalTrain.show()
		
	def remove_protocol_from_list(self):

		current_item = self.protocol_list.currentRow()
		if current_item > -1:
			del self.protocols[current_item]
			self.protocol_list.takeItem(current_item)

	def open_config_distance_threshold_neighborhood(self):

		self.ConfigNeigh = ConfigNeighborhoods(parent_window=self,
											   neighborhood_type='distance_threshold',
											   neighborhood_parameter_name='threshold distance',
											   )
		self.ConfigNeigh.show()

	def open_config_contact_neighborhood(self):

		self.ConfigNeigh = ConfigNeighborhoods(parent_window=self,
											   neighborhood_type='mask_contact',
											   neighborhood_parameter_name='tolerance contact distance',
											   )
		self.ConfigNeigh.show()

	def enable_signal_model_list(self):
		if self.signal_analysis_action.isChecked():
			self.pair_signal_models_list.setEnabled(True)
		else:
			self.pair_signal_models_list.setEnabled(False)

	def process_neighborhood(self):

		if self.parent_window.well_list.currentText()=="*":
			self.well_index = np.linspace(0,len(self.wells)-1,len(self.wells),dtype=int)
		else:
			self.well_index = [self.parent_window.well_list.currentIndex()]
			print(f"Processing well {self.parent_window.well_list.currentText()}...")

		# self.freeze()
		# QApplication.setOverrideCursor(Qt.WaitCursor)

		loop_iter=0

		if self.parent_window.position_list.currentText()=="*":
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Question)
			msgBox.setText("If you continue, all positions will be processed.\nDo you want to proceed?")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.No:
				return None

		for w_idx in self.well_index:

			pos = self.parent_window.positions[w_idx]
			if self.parent_window.position_list.currentText()=="*":
				pos_indices = np.linspace(0,len(pos)-1,len(pos),dtype=int)
				print("Processing all positions...")
			else:
				pos_indices = natsorted([pos.index(self.parent_window.position_list.currentText())])
				print(f"Processing position {self.parent_window.position_list.currentText()}...")

			well = self.parent_window.wells[w_idx]

			for pos_idx in pos_indices:

				self.pos = natsorted(glob(well+f"{os.path.split(well)[-1].replace('W','').replace(os.sep,'')}*{os.sep}"))[pos_idx]
				print(f"Position {self.pos}...\nLoading stack movie...")

				if not os.path.exists(self.pos + 'output' + os.sep):
					os.mkdir(self.pos + 'output' + os.sep)
				if not os.path.exists(self.pos + os.sep.join(['output','tables'])+os.sep):
					os.mkdir(self.pos + os.sep.join(['output','tables'])+os.sep)

				if self.neigh_action.isChecked():
					for protocol in self.protocols:

						if protocol['neighborhood_type']=='distance_threshold':

							compute_neighborhood_at_position(self.pos,
															protocol['distance'],
															population=protocol['population'],
															theta_dist=None,
															img_shape=(self.parent_window.shape_x,self.parent_window.shape_y),
															return_tables=False,
															clear_neigh=protocol['clear_neigh'],
															event_time_col=protocol['event_time_col'],
															neighborhood_kwargs=protocol['neighborhood_kwargs'],
															)

						elif protocol['neighborhood_type']=='mask_contact':

							compute_contact_neighborhood_at_position(self.pos,
														protocol['distance'],
														population=protocol['population'],
														theta_dist=None,
														img_shape=(self.parent_window.shape_x,self.parent_window.shape_y),
														return_tables=False,
														clear_neigh=protocol['clear_neigh'],
														event_time_col=protocol['event_time_col'],
														neighborhood_kwargs=protocol['neighborhood_kwargs'],
														)
				if self.measure_pairs_action.isChecked():
					rel_measure_at_position(self.pos)

				if self.signal_analysis_action.isChecked():
					
					# df_targets = get_position_pickle(self.pos, population='targets')
					# df_effectors = get_position_pickle(self.pos, population='effectors')
					# self.dataframes = {
					# 	'targets': df_targets,
					# 	'effectors': df_effectors,
					# }

					# df_pairs = get_position_table(self.pos, population='pairs')

					# # Need to identify expected reference / neighbor tables
					# model_path = locate_signal_model(self.pair_signal_models_list.currentText(), pairs=True)
					# print(f'Looking for model in {model_path}...')
					# complete_path = model_path
					# complete_path = rf"{complete_path}"
					# model_config_path = os.sep.join([complete_path, 'config_input.json'])
					# model_config_path = rf"{model_config_path}"
					# f = open(model_config_path)
					# model_config_path = json.load(f)

					# reference_population = model_config_path['reference_population']
					# neighbor_population = model_config_path['neighbor_population']

					# analyze_pair_signals(df_pairs, self.dataframes[reference_population], self.dataframes[neighbor_population], model=self.pair_signal_models_list.currentText())
					analyze_pair_signals_at_position(self.pos, self.pair_signal_models_list.currentText(), use_gpu=self.parent_window.parent_window.use_gpu)
		self.parent_window.update_position_options()
		print('Done.')


	def check_signals2(self):

		test = self.parent_window.locate_selected_position()
		if test:
			self.SignalAnnotator2 = SignalAnnotator2(self)
			self.SignalAnnotator2.show()

	def check_measurements2(self):

		test = self.parent_window.locate_selected_position()
		if test:
			self.MeasurementAnnotator2 = MeasureAnnotator2(self)
			self.MeasurementAnnotator2.show()



class PreprocessingPanel(QFrame, Styles):

	def __init__(self, parent_window):

		super().__init__()
		self.parent_window = parent_window
		self.exp_channels = self.parent_window.exp_channels
		self.exp_dir = self.parent_window.exp_dir
		self.wells = np.array(self.parent_window.wells,dtype=str)
		exp_config = self.exp_dir + "config.ini"
		self.channel_names, self.channels = extract_experiment_channels(exp_config)
		self.channel_names = np.array(self.channel_names)
		self.background_correction = []
		self.onlyFloat = QDoubleValidator()
		self.onlyInt = QIntValidator()

		self.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
		self.grid = QGridLayout(self)

		self.generate_header()

	def generate_header(self):

		"""
		Read the mode and prepare a collapsable block to process a specific cell population.

		"""

		panel_title = QLabel(f"PREPROCESSING")
		panel_title.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")

		self.grid.addWidget(panel_title, 0, 0, 1, 4, alignment=Qt.AlignCenter)

		self.select_all_btn = QPushButton()
		self.select_all_btn.setIcon(icon(MDI6.checkbox_blank_outline,color="black"))
		self.select_all_btn.setIconSize(QSize(20, 20))
		self.all_ticked = False
		#self.select_all_btn.clicked.connect(self.tick_all_actions)
		self.select_all_btn.setStyleSheet(self.button_select_all)
		self.grid.addWidget(self.select_all_btn, 0, 0, 1, 4, alignment=Qt.AlignLeft)
		#self.to_disable.append(self.all_tc_actions)

		self.collapse_btn = QPushButton()
		self.collapse_btn.setIcon(icon(MDI6.chevron_down,color="black"))
		self.collapse_btn.setIconSize(QSize(25, 25))
		self.collapse_btn.setStyleSheet(self.button_select_all)
		self.grid.addWidget(self.collapse_btn, 0, 0, 1, 4, alignment=Qt.AlignRight)

		self.populate_contents()

		self.grid.addWidget(self.ContentsFrame, 1, 0, 1, 4, alignment=Qt.AlignTop)
		self.collapse_btn.clicked.connect(lambda: self.ContentsFrame.setHidden(not self.ContentsFrame.isHidden()))
		self.collapse_btn.clicked.connect(self.collapse_advanced)
		self.ContentsFrame.hide()

	def collapse_advanced(self):

		effector_open = not self.parent_window.ProcessEffectors.ContentsFrame.isHidden()
		targets_open = not self.parent_window.ProcessTargets.ContentsFrame.isHidden()
		interactions_open = not self.parent_window.NeighPanel.ContentsFrame.isHidden()
		preprocessing_open = not self.parent_window.PreprocessingPanel.ContentsFrame.isHidden()
		is_open = np.array([effector_open, targets_open, interactions_open, preprocessing_open])

		if self.ContentsFrame.isHidden():
			self.collapse_btn.setIcon(icon(MDI6.chevron_down,color="black"))
			self.collapse_btn.setIconSize(QSize(20, 20))
			if len(is_open[is_open])==0:
				self.parent_window.scroll.setMinimumHeight(int(550))
				self.parent_window.adjustSize()
		else:
			self.collapse_btn.setIcon(icon(MDI6.chevron_up,color="black"))
			self.collapse_btn.setIconSize(QSize(20, 20))
			self.parent_window.scroll.setMinimumHeight(min(int(930), int(0.9*self.parent_window.screen_height)))

	def populate_contents(self):

		self.ContentsFrame = QFrame()
		self.grid_contents = QGridLayout(self.ContentsFrame)

		self.model_free_correction_layout = BackgroundModelFreeCorrectionLayout(self)
		self.fit_correction_layout = BackgroundFitCorrectionLayout(self)

		self.protocol_layout = ProtocolDesignerLayout(parent_window=self,
											  tab_layouts=[self.fit_correction_layout, self.model_free_correction_layout],
											  tab_names=['Fit', 'Model-free'],
											  title='BACKGROUND CORRECTION',
											  list_title='Corrections to apply:')
		
		self.help_background_btn = QPushButton()
		self.help_background_btn.setIcon(icon(MDI6.help_circle,color=self.help_color))
		self.help_background_btn.setIconSize(QSize(20, 20))
		self.help_background_btn.clicked.connect(self.help_background)
		self.help_background_btn.setStyleSheet(self.button_select_all)
		self.help_background_btn.setToolTip("Help.")

		self.protocol_layout.title_layout.addWidget(self.help_background_btn, 5, alignment=Qt.AlignRight)

		self.grid_contents.addLayout(self.protocol_layout,0,0,1,4)
		self.submit_preprocessing_btn = QPushButton("Submit")
		self.submit_preprocessing_btn.setStyleSheet(self.button_style_sheet_2)
		self.submit_preprocessing_btn.clicked.connect(self.launch_preprocessing)
		self.grid_contents.addWidget(self.submit_preprocessing_btn, 1,0,1,4)


	def launch_preprocessing(self):

		msgBox1 = QMessageBox()
		msgBox1.setIcon(QMessageBox.Question)
		msgBox1.setText("Do you want to apply the preprocessing\nto all wells and positions?")
		msgBox1.setWindowTitle("Selection")
		msgBox1.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
		returnValue = msgBox1.exec()
		if returnValue == QMessageBox.Cancel:
			return None
		elif returnValue == QMessageBox.Yes:
			self.parent_window.well_list.setCurrentIndex(self.parent_window.well_list.count()-1)
			self.parent_window.position_list.setCurrentIndex(0)
		elif returnValue == QMessageBox.No:
			msgBox2 = QMessageBox()
			msgBox2.setIcon(QMessageBox.Question)
			msgBox2.setText("Do you want to apply the preprocessing\nto the positions selected at the top only?")
			msgBox2.setWindowTitle("Selection")
			msgBox2.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
			returnValue = msgBox2.exec()
			if returnValue == QMessageBox.Cancel:
				return None
			if returnValue == QMessageBox.No:
				return None

		print('Proceed with correction...')

		if self.parent_window.well_list.currentText()=='*':
			well_option = "*"
		else:
			well_option = self.parent_window.well_list.currentIndex()

		if self.parent_window.position_list.currentText()=='*':
			pos_option = "*"
		else:
			pos_option = self.parent_window.position_list.currentIndex()-1


		for k,correction_protocol in enumerate(self.protocol_layout.protocols):

			movie_prefix = None
			export_prefix = 'Corrected'
			if k>0:
				# switch source stack to cumulate multi-channel preprocessing
				movie_prefix = 'Corrected'
				export_prefix = None

			if correction_protocol['correction_type']=='model-free':

				correct_background_model_free(self.exp_dir, 
								   well_option=well_option,
								   position_option=pos_option,
								   export = True,
								   return_stacks=False,
								   show_progress_per_well = True,
								   show_progress_per_pos = True,
								   movie_prefix = movie_prefix,
								   export_prefix = export_prefix,
								   **correction_protocol,
								)
			
			elif correction_protocol['correction_type']=='fit':

				correct_background_model(self.exp_dir,
								   well_option=well_option,
								   position_option=pos_option,
								   export= True,
								   return_stacks=False,
								   show_progress_per_well = True,
								   show_progress_per_pos = True,
								   movie_prefix = movie_prefix,
								   export_prefix = export_prefix,
								   **correction_protocol,
								)
		print('Done.')


	def locate_image(self):

		"""
		Load the first frame of the first movie found in the experiment folder as a sample.
		"""

		movies = glob(self.parent_window.pos + os.sep.join(['movie', f"{self.parent_window.movie_prefix}*.tif"]))

		if len(movies) == 0:
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("Please select a position containing a movie...")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				self.current_stack = None
				return None
		else:
			self.current_stack = movies[0]

	def help_background(self):

		"""
		Helper to choose a proper cell population structure.
		"""

		dict_path = os.sep.join([get_software_location(),'celldetective','gui','help','preprocessing.json'])

		with open(dict_path) as f:
			d = json.load(f)

		suggestion = help_generic(d)
		if isinstance(suggestion, str):
			print(f"{suggestion=}")
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Information)
			msgBox.setTextFormat(Qt.RichText)
			msgBox.setText(suggestion)
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None			



from PyQt5.QtWidgets import QMainWindow, QApplication, QMessageBox, QHBoxLayout, QFileDialog, QVBoxLayout, QScrollArea, QCheckBox, QGridLayout, QLabel, QLineEdit, QPushButton, QWidget
from PyQt5.QtGui import QIntValidator, QDoubleValidator
from celldetective.gui.gui_utils import center_window, help_generic
from celldetective.gui.styles import Styles
from celldetective.utils import get_software_location
import json

from superqt import QLabeledSlider
from PyQt5.QtCore import Qt, QSize
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
from configparser import ConfigParser
import os
from functools import partial
import numpy as np
from celldetective.gui import Styles

class ConfigNewExperiment(QMainWindow, Styles):

	def __init__(self, parent_window=None):
		
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("New experiment")
		center_window(self)
		self.setFixedWidth(500)
		self.setMaximumHeight(int(0.8*self.parent_window.screen_height))
		self.onlyFloat = QDoubleValidator()
		
		self.newExpFolder = str(QFileDialog.getExistingDirectory(self, 'Select directory'))
		self.populate_widget()

	def populate_widget(self):
		
		# Create button widget and layout
		self.scroll_area = QScrollArea(self)
		button_widget = QWidget()
		self.grid = QGridLayout()
		button_widget.setLayout(self.grid)

		self.grid.setContentsMargins(30,30,30,30)
		self.grid.addWidget(QLabel("Folder:"), 0, 0, 1, 3)
		self.supFolder = QLineEdit()
		self.supFolder.setAlignment(Qt.AlignLeft)	
		self.supFolder.setEnabled(True)
		self.supFolder.setText(self.newExpFolder)
		self.grid.addWidget(self.supFolder, 1, 0, 1, 1)

		self.browse_button = QPushButton("Browse...")
		self.browse_button.clicked.connect(self.browse_experiment_folder)
		self.browse_button.setIcon(icon(MDI6.folder, color="white"))
		self.browse_button.setStyleSheet(self.button_style_sheet)
		#self.browse_button.setIcon(QIcon_from_svg(abs_path+f"/icons/browse.svg", color='white'))
		self.grid.addWidget(self.browse_button, 1, 1, 1, 1)

		self.grid.addWidget(QLabel("Experiment name:"), 2, 0, 1, 3)
		
		self.expName = QLineEdit()
		self.expName.setPlaceholderText('folder_name_for_the_experiment')
		self.expName.setAlignment(Qt.AlignLeft)	
		self.expName.setEnabled(True)
		self.expName.setFixedWidth(400)
		self.expName.setText("Untitled_Experiment")
		self.grid.addWidget(self.expName, 3, 0, 1, 3)

		self.generate_movie_settings()
		self.grid.addLayout(self.ms_grid,29,0,1,3)

		self.generate_channel_params_box()
		self.grid.addLayout(self.channel_grid,30,0,1,3)

		self.validate_button = QPushButton("Submit")
		self.validate_button.clicked.connect(self.create_config)
		self.validate_button.setStyleSheet(self.button_style_sheet)
		#self.validate_button.setIcon(QIcon_from_svg(abs_path+f"/icons/process.svg", color='white'))

		self.grid.addWidget(self.validate_button, 31, 0, 1, 3, alignment = Qt.AlignBottom)		
		button_widget.adjustSize()

		self.scroll_area.setAlignment(Qt.AlignCenter)
		self.scroll_area.setWidget(button_widget)
		self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
		self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
		self.scroll_area.setWidgetResizable(True)
		self.setCentralWidget(self.scroll_area)
		self.show()

		QApplication.processEvents()
		self.adjustScrollArea()

	def adjustScrollArea(self):
		
		"""
		Auto-adjust scroll area to fill space 
		(from https://stackoverflow.com/questions/66417576/make-qscrollarea-use-all-available-space-of-qmainwindow-height-axis)
		"""

		step = 5
		while self.scroll_area.verticalScrollBar().isVisible() and self.height() < self.maximumHeight():
			self.resize(self.width(), self.height() + step)

	def generate_movie_settings(self):
		
		"""
		Parameters related to the movie parameters
		"""

		onlyInt = QIntValidator()
		onlyInt.setRange(0, 100000)

		self.ms_grid = QGridLayout()
		self.ms_grid.setContentsMargins(21,30,20,30)

		ms_lbl = QLabel("MOVIE SETTINGS")
		ms_lbl.setStyleSheet("""
			font-weight: bold;
			""")
		self.ms_grid.addWidget(ms_lbl, 0,0,1,3, alignment=Qt.AlignCenter)

		self.number_of_wells = QLabel("Number of wells:")
		self.ms_grid.addWidget(self.number_of_wells, 1, 0, 1, 3)

		self.help_btn = QPushButton()
		self.help_btn.setIcon(icon(MDI6.help_circle,color=self.help_color))
		self.help_btn.setIconSize(QSize(20, 20))
		self.help_btn.clicked.connect(self.help_structure)
		self.help_btn.setStyleSheet(self.button_select_all)
		self.help_btn.setToolTip("Help.")
		self.ms_grid.addWidget(self.help_btn, 1, 0, 1, 3, alignment=Qt.AlignRight)


		self.SliderWells = QLabeledSlider(Qt.Horizontal, self)
		self.SliderWells.setMinimum(1)
		self.SliderWells.setMaximum(32)
		self.ms_grid.addWidget(self.SliderWells, 2, 0, 1, 3, alignment=Qt.AlignTop)

		self.number_of_positions = QLabel("Number of positions per well:")
		self.ms_grid.addWidget(self.number_of_positions, 3, 0, 1, 3)

		self.SliderPos = QLabeledSlider(Qt.Horizontal, self)
		self.SliderPos.setMinimum(1)
		self.SliderPos.setMaximum(50)

		self.ms_grid.addWidget(self.SliderPos, 4, 0, 1, 3)

		self.ms_grid.addWidget(QLabel("Calibration from pixel to µm:"), 5, 0, 1, 3)
		self.PxToUm_field = QLineEdit()
		self.PxToUm_field.setValidator(self.onlyFloat)
		self.PxToUm_field.setPlaceholderText('1 px = XXX µm')
		self.PxToUm_field.setAlignment(Qt.AlignLeft)	
		self.PxToUm_field.setEnabled(True)
		self.PxToUm_field.setFixedWidth(400)
		self.PxToUm_field.setText("1,0")
		self.ms_grid.addWidget(self.PxToUm_field, 6, 0, 1, 3)

		self.ms_grid.addWidget(QLabel("Calibration from frame to minutes:"), 7, 0, 1, 3)
		self.FrameToMin_field = QLineEdit()
		self.FrameToMin_field.setAlignment(Qt.AlignLeft)	
		self.FrameToMin_field.setEnabled(True)
		self.FrameToMin_field.setFixedWidth(400)
		self.FrameToMin_field.setValidator(self.onlyFloat)
		self.FrameToMin_field.setPlaceholderText('1 frame = XXX min')
		self.FrameToMin_field.setText("1,0")
		self.ms_grid.addWidget(self.FrameToMin_field, 8, 0, 1, 3)

		self.movie_length = QLabel("Number of frames:")
		self.movie_length.setToolTip('Optional: depending on how the movies are encoded, the automatic extraction of the number of frames can be difficult.\nThe software will then rely on this value.')
		self.ms_grid.addWidget(self.movie_length,9, 0, 1, 3)
		self.MovieLengthSlider = QLabeledSlider(Qt.Horizontal, self)
		self.MovieLengthSlider.setMinimum(2)
		self.MovieLengthSlider.setMaximum(128)
		self.ms_grid.addWidget(self.MovieLengthSlider, 10, 0, 1, 3)

		self.prefix_lbl = QLabel("Prefix for the movies:")
		self.prefix_lbl.setToolTip('The stack file name must start with this prefix to be properly loaded.')
		self.ms_grid.addWidget(self.prefix_lbl, 11, 0, 1, 3)
		self.movie_prefix_field = QLineEdit()
		self.movie_prefix_field.setAlignment(Qt.AlignLeft)	
		self.movie_prefix_field.setEnabled(True)
		self.movie_prefix_field.setFixedWidth(400)
		self.movie_prefix_field.setText("")
		self.ms_grid.addWidget(self.movie_prefix_field, 12, 0, 1, 3)

		self.ms_grid.addWidget(QLabel("X shape in pixels:"), 13, 0, 1, 3)
		self.shape_x_field = QLineEdit()
		self.shape_x_field.setValidator(onlyInt)
		self.shape_x_field.setAlignment(Qt.AlignLeft)	
		self.shape_x_field.setEnabled(True)
		self.shape_x_field.setFixedWidth(400)
		self.shape_x_field.setText("2048")
		self.ms_grid.addWidget(self.shape_x_field, 14, 0, 1, 3)

		self.ms_grid.addWidget(QLabel("Y shape in pixels:"), 15, 0, 1, 3)
		self.shape_y_field = QLineEdit()
		self.shape_y_field.setValidator(onlyInt)
		self.shape_y_field.setAlignment(Qt.AlignLeft)	
		self.shape_y_field.setEnabled(True)
		self.shape_y_field.setFixedWidth(400)
		self.shape_y_field.setText("2048")
		self.ms_grid.addWidget(self.shape_y_field, 16, 0, 1, 3)

	def help_structure(self):

		"""
		Helper to choose an experiment structure.
		"""

		dict_path = os.sep.join([get_software_location(),'celldetective','gui','help','exp-structure.json'])

		with open(dict_path) as f:
			d = json.load(f)

		suggestion = help_generic(d)
		if isinstance(suggestion, str):
			print(f"{suggestion=}")
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Information)
			msgBox.setTextFormat(Qt.RichText)
			msgBox.setText(suggestion+"\nSee <a href='https://celldetective.readthedocs.io/en/latest/get-started.html#data-organization'>the docs</a> for more information.")
			msgBox.setWindowTitle("Info")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None		

	def generate_channel_params_box(self):

		"""
		Parameters related to the movie channels
		Rewrite all of it

		"""

		self.channel_grid = QGridLayout()
		self.channel_grid.setContentsMargins(21,30,20,30)

		channel_lbl = QLabel("CHANNELS")
		channel_lbl.setStyleSheet("""
			font-weight: bold;
			""")
		self.channel_grid.addWidget(channel_lbl, 0,0,1,3, alignment=Qt.AlignCenter)


		self.channels = ['brightfield', "live nuclei channel\n(Hoechst, NucSpot®)", "dead nuclei channel\n(PI)", "effector fluorescence\n(CFSE)",
						"adhesion\n(RICM, IRM)", "fluorescence miscellaneous 1\n(LAMP-1, Actin...)", "fluorescence miscellaneous 2\n(LAMP-1, Actin...)"]
		self.channel_mapping = ['brightfield_channel', 'live_nuclei_channel', 'dead_nuclei_channel', 'effector_fluo_channel',
								'adhesion_channel', 'fluo_channel_1', 'fluo_channel_2']
		self.checkBoxes = [QCheckBox() for i in range(len(self.channels))]
		self.sliders = [QLabeledSlider(Qt.Orientation.Horizontal) for i in range(len(self.channels))]

		for i in range(len(self.channels)):

			self.checkBoxes[i].setText(self.channels[i])
			self.checkBoxes[i].toggled.connect(partial(self.show_slider, i))

			self.channel_grid.addWidget(self.checkBoxes[i], i+1, 0, 1, 1)

			self.sliders[i].setMinimum(0)
			self.sliders[i].setMaximum(len(self.channels))
			self.sliders[i].setEnabled(False)

			self.channel_grid.addWidget(self.sliders[i], i+1, 1, 1, 2, alignment = Qt.AlignRight)

		# Add channel button
		self.addChannelBtn = QPushButton('Add channel')
		self.addChannelBtn.setIcon(icon(MDI6.plus,color="white"))
		self.addChannelBtn.setIconSize(QSize(25, 25))
		self.addChannelBtn.setStyleSheet(self.button_style_sheet)
		self.addChannelBtn.clicked.connect(self.add_custom_channel)
		self.channel_grid.addWidget(self.addChannelBtn, 1000, 0, 1, 1)

	def add_custom_channel(self):

		self.CustomChannelWidget = QWidget()
		self.CustomChannelWidget.setWindowTitle("Custom channel")
		layout = QVBoxLayout()
		self.CustomChannelWidget.setLayout(layout)

		self.name_le = QLineEdit('custom_channel')
		hbox = QHBoxLayout()
		hbox.addWidget(QLabel('channel name: '), 33)
		hbox.addWidget(self.name_le, 66)
		layout.addLayout(hbox)

		self.createBtn = QPushButton('create')
		self.createBtn.setStyleSheet(self.button_style_sheet)
		self.createBtn.clicked.connect(self.write_custom_channel)
		layout.addWidget(self.createBtn)
		center_window(self.CustomChannelWidget)
		self.CustomChannelWidget.show()

		print('new channel added')

	def write_custom_channel(self):
		
		self.new_channel_name = self.name_le.text()
		name_map = self.new_channel_name
		name_map = name_map.replace('_channel','')
		name_map = name_map.replace('channel','')
		name_map = name_map.replace(' ','')
		if not name_map.endswith('_channel'):
			name_map += '_channel'

		self.channels.append(self.new_channel_name)
		self.channel_mapping.append(name_map)
		self.checkBoxes.append(QCheckBox())
		self.sliders.append(QLabeledSlider(Qt.Orientation.Horizontal))
		self.CustomChannelWidget.close()

		self.checkBoxes[-1].setText(self.channels[-1])
		self.checkBoxes[-1].toggled.connect(partial(self.show_slider, len(self.channels)-1))
		self.channel_grid.addWidget(self.checkBoxes[-1], len(self.channels)+1, 0, 1, 1)

		self.sliders[-1].setMinimum(0)
		for i in range(len(self.channels)):
			self.sliders[i].setMaximum(len(self.channels))
		self.sliders[-1].setEnabled(False)
		self.channel_grid.addWidget(self.sliders[-1], len(self.channels)+1, 1, 1, 2, alignment = Qt.AlignRight)

	def show_slider(self, index):
		if self.checkBoxes[index].isChecked():
			self.sliders[index].setEnabled(True)
		else:
			self.sliders[index].setEnabled(False)

	def browse_experiment_folder(self):

		"""
		Set a new base directory.
		"""

		self.newExpFolder = str(QFileDialog.getExistingDirectory(self, 'Select directory'))
		self.supFolder.setText(self.newExpFolder)

	def create_config(self):

		"""
		Create the folder tree, the config, issue a warning if the experiment folder already exists.
		"""

		channel_indices = []
		for i in range(len(self.channels)):
			if self.checkBoxes[i].isChecked():
				channel_indices.append(self.sliders[i].value())
		if not channel_indices:
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("Please set at least one channel before proceeding.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

		if len(channel_indices) != len(set(channel_indices)):
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("Some channel indices are repeated. Please check your configuration.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

		sorted_list = set(channel_indices)
		expected_list = set(np.arange(max(sorted_list)+1))
		print(sorted_list, expected_list, sorted_list==expected_list)

		if not sorted_list==expected_list:
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("There is a gap in your channel indices. Please check your configuration.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None			
		
		try:
			
			folder = self.supFolder.text()
			folder = folder.replace('\\','/')
			folder = rf"{folder}"

			name = str(self.expName.text())
			name = name.replace(' ','')

			self.directory = os.sep.join([folder,name])
			os.mkdir(self.directory)
			os.chdir(self.directory)
			self.create_subfolders()
			self.annotate_wells()

		except FileExistsError:
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("This experiment already exists... Please select another name.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

	def create_subfolders(self):

		"""
		Create the well folders and position folders within the wells.
		"""
		
		self.nbr_wells = self.SliderWells.value()
		self.nbr_positions = self.SliderPos.value()
		for k in range(self.nbr_wells):
			well_name = f"W{k+1}"+os.sep
			os.mkdir(well_name)
			for p in range(self.nbr_positions):
				position_name = well_name+f"{k+1}0{p}"+os.sep
				os.mkdir(position_name)
				os.mkdir(position_name+os.sep+"movie"+os.sep)

	def annotate_wells(self):
		self.w = SetupConditionLabels(self, self.nbr_wells)
		self.w.show()


	def create_config_file(self):

		"""

		Write all user input parameters to a configuration file associated to an experiment.
		"""

		config = ConfigParser()

		# add a new section and some values
		config.add_section('MovieSettings')
		config.set('MovieSettings', 'PxToUm', self.PxToUm_field.text().replace(',','.'))
		config.set('MovieSettings', 'FrameToMin', self.FrameToMin_field.text().replace(',','.'))
		config.set('MovieSettings', 'len_movie', str(self.MovieLengthSlider.value()))
		config.set('MovieSettings', 'shape_x', self.shape_x_field.text())
		config.set('MovieSettings', 'shape_y', self.shape_y_field.text())
		config.set('MovieSettings', 'movie_prefix', self.movie_prefix_field.text())

		config.add_section('Channels')
		for i in range(len(self.channels)):
			if self.checkBoxes[i].isChecked():
				config.set('Channels', self.channel_mapping[i], str(self.sliders[i].value()))
			else:
				config.set('Channels', self.channel_mapping[i], "nan")

		config.add_section('Labels')
		config.set('Labels', 'cell_types', self.cell_types)
		config.set('Labels', 'antibodies', self.antibodies)
		config.set('Labels', 'concentrations', self.concentrations)
		config.set('Labels', 'pharmaceutical_agents', self.pharmaceutical_agents)

		# save to a file
		with open('config.ini', 'w') as configfile:
			config.write(configfile)

		self.parent_window.set_experiment_path(self.directory)
		print(f'New experiment successfully configured in folder {self.directory}...')
		self.close()

class SetupConditionLabels(QWidget, Styles):
	def __init__(self, parent_window, n_wells):
		super().__init__()
		self.parent_window = parent_window
		self.n_wells = n_wells
		self.setWindowTitle("Well conditions")
		self.layout = QVBoxLayout()
		self.layout.setContentsMargins(30,30,30,30)
		self.setLayout(self.layout)
		self.onlyFloat = QDoubleValidator()
		self.populate()
		center_window(self)

	def populate(self):

		self.cell_type_cbs = [QLineEdit() for i in range(self.n_wells)]
		self.antibodies_cbs = [QLineEdit() for i in range(self.n_wells)]
		self.concentrations_cbs = [QLineEdit() for i in range(self.n_wells)]
		self.pharmaceutical_agents_cbs = [QLineEdit() for i in range(self.n_wells)]

		for i in range(self.n_wells):
			hbox = QHBoxLayout()
			hbox.setContentsMargins(15,5,15,5)
			hbox.addWidget(QLabel(f'well {i+1}'), 5, alignment=Qt.AlignLeft)
			hbox.addWidget(QLabel('cell type: '), 5)
			hbox.addWidget(self.cell_type_cbs[i], 10)
			self.cell_type_cbs[i].setPlaceholderText('e.g. T-cell, NK')

			hbox.addWidget(QLabel('antibody: '), 5)
			hbox.addWidget(self.antibodies_cbs[i], 10)
			self.antibodies_cbs[i].setPlaceholderText('e.g. anti-CD4')
			
			hbox.addWidget(QLabel('concentration: '), 5)
			hbox.addWidget(self.concentrations_cbs[i], 10)
			self.concentrations_cbs[i].setPlaceholderText('e.g. 100 (pM)')
			self.concentrations_cbs[i].setValidator(self.onlyFloat)
			
			hbox.addWidget(QLabel('pharmaceutical agents: '), 5)
			hbox.addWidget(self.pharmaceutical_agents_cbs[i], 10)
			self.pharmaceutical_agents_cbs[i].setPlaceholderText('e.g. dextran')

			self.layout.addLayout(hbox)

		btn_hbox = QHBoxLayout()
		btn_hbox.setContentsMargins(0,20,0,0)
		self.skip_btn = QPushButton('Skip')
		self.skip_btn.setStyleSheet(self.button_style_sheet_2)
		self.skip_btn.clicked.connect(self.set_default_values)
		btn_hbox.addWidget(self.skip_btn)

		self.submit_btn = QPushButton('Submit')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.set_user_values)
		btn_hbox.addWidget(self.submit_btn)
		self.layout.addLayout(btn_hbox)

	def set_default_values(self):
		
		for i in range(self.n_wells):
			self.cell_type_cbs[i].setText(str(i))
			self.antibodies_cbs[i].setText(str(i))
			self.concentrations_cbs[i].setText(str(i))
			self.pharmaceutical_agents_cbs[i].setText(str(i))
		self.set_attributes()
		self.parent_window.create_config_file()
		self.close()

	def set_user_values(self):
		for i in range(self.n_wells):
			if self.cell_type_cbs[i].text()=='':
				self.cell_type_cbs[i].setText(str(i))
			if self.antibodies_cbs[i].text()=='':
				self.antibodies_cbs[i].setText(str(i))
			if self.concentrations_cbs[i].text()=='':
				self.concentrations_cbs[i].setText(str(i))
			if self.pharmaceutical_agents_cbs[i].text()=='':
				self.pharmaceutical_agents_cbs[i].setText(str(i))
		self.set_attributes()
		self.parent_window.create_config_file()
		self.close()

	def set_attributes(self):

		cell_type_text = [c.text() for c in self.cell_type_cbs]
		self.parent_window.cell_types = ','.join(cell_type_text)

		antibodies_text = [c.text() for c in self.antibodies_cbs]
		self.parent_window.antibodies = ','.join(antibodies_text)

		concentrations_text = [c.text() for c in self.concentrations_cbs]
		self.parent_window.concentrations = ','.join(concentrations_text)

		pharamaceutical_text = [c.text() for c in self.pharmaceutical_agents_cbs]
		self.parent_window.pharmaceutical_agents = ','.join(pharamaceutical_text)





